import pexpect
import pytest
from conftest import ShellTest
from test_helper import KEY_ACTION


@pytest.mark.parametrize(
    "esc",
    [
        "\x1bb",
        "\x1b[1;5D",
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_move_word_left(test, esc) -> None:
    test.send_statement("RETURN AS a;")
    test.send_statement(esc * 3)
    test.send_finished_statement(' "databases rule"\x1b[F\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "esc",
    [
        "\x1bf",
        "\x1b[1;5C",
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_move_word_right(test, esc) -> None:
    test.send_statement('RETURN "databases a;')
    test.send_control_statement(KEY_ACTION.CTRL_A.value)  # move cursor to front of line
    test.send_statement(esc * 3)
    test.send_finished_statement(' rule" AS\x1b[F\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "esc",
    [
        "\x1b[H",
        "\x1b[1~",
        "\x1bOH",
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_move_home(test, esc) -> None:
    test.send_statement('ETURN "databases rule" AS a;')
    test.send_statement(esc)
    test.send_finished_statement("R\x1b[F\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_delete(test) -> None:
    test.send_statement('RETURN "databases rule" AS a;abc')
    for _ in range(3):
        test.send_control_statement(KEY_ACTION.CTRL_B.value)
    test.send_statement("\x1b[3~" * 4)
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize(
    "esc",
    [
        "\x1b[F",
        "\x1b[4~",
        "\x1bOF",
    ],
)
@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_move_end(test, esc) -> None:
    test.send_statement('RETURN "databases rule" AS a')
    test.send_control_statement(KEY_ACTION.CTRL_A.value)  # move cursor to front of line
    test.send_statement(esc)
    test.send_finished_statement(";\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_next_history_singleline(test) -> None:
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_statement("\x1b[A")
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


def test_next_history_multiline(temp_db) -> None:
    # multiline up arrow will move up if not on top line
    test = ShellTest().add_argument(temp_db)
    test.start()
    test.send_finished_statement('RETURN "shell is fun" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 shell is fun \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "kuzu is cool"\r AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_statement("\x1b[A\x1b[A\x1b[A")
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0

    # singleline up arrow will move up regardless
    test = ShellTest().add_argument(temp_db)
    test.start()
    test.send_finished_statement(':singleline\n')
    test.send_finished_statement('RETURN "shell is fun" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 shell is fun \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "kuzu is cool"\r AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_statement("\x1b[A\x1b[A\x1b[A")
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 shell is fun \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_prev_history_singleline(test) -> None:
    test.send_finished_statement('RETURN "kuzu is cool" AS b;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_P.value)  # move up in history
    test.send_control_statement(KEY_ACTION.CTRL_P.value)  # move up in history
    test.send_statement("\x1b[B")
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_prev_history_multiline(test) -> None:
    # multiline down arrow will move down if not on bottom line
    test.send_finished_statement('RETURN "kuzu is cool"\r AS b;\r')
    assert test.shell_process.expect_exact(["\u2502 kuzu is cool \u2502", pexpect.EOF]) == 0
    test.send_finished_statement('RETURN "databases rule" AS a;\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0
    test.send_control_statement(KEY_ACTION.CTRL_P.value)  # move up in history
    test.send_control_statement(KEY_ACTION.CTRL_P.value)  # move up in history
    test.send_control_statement(KEY_ACTION.CTRL_P.value)  # move up line
    test.send_statement("\x1b[B")
    test.send_statement("\x1b[B")
    test.send_finished_statement(KEY_ACTION.ENTER.value)
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_move_left(test) -> None:
    test.send_statement(" a;")
    test.send_statement("\x1b[D" * 4)
    test.send_finished_statement('RETURN "databases rule" AS\x1b[F\r')
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0


@pytest.mark.parametrize('test', ['multiline', 'singleline'], indirect=True)
def test_move_right(test) -> None:
    test.send_statement('RETURN "databases rule" AS a')
    for _ in range(3):
        test.send_control_statement(KEY_ACTION.CTRL_B.value)  # move cursor to the left
    test.send_statement("\x1b[C" * 4)
    test.send_finished_statement(";\r")
    assert test.shell_process.expect_exact(["\u2502 databases rule \u2502", pexpect.EOF]) == 0



#!/usr/bin/env python3
import argparse
import io
import json
import multiprocessing
import os
import subprocess
import sys


class LSPClient:
    def __init__(self, *, compile_commands_dir=None, jobs=None, verbose=False):
        args = ["clangd", "-j", str(jobs)]
        if jobs is None:
            jobs = multiprocessing.cpu_count()

        if compile_commands_dir is not None:
            args += ["--compile-commands-dir", compile_commands_dir]

        self.id = 0
        self.child = subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            # Let clangd inherit our stderr, or suppress it entirely.
            stderr=None if verbose else subprocess.DEVNULL,
        )
        self.stdin = io.TextIOWrapper(self.child.stdin, newline="\r\n")
        self.stdout = io.TextIOWrapper(self.child.stdout, newline="\r\n")

    def request(self, method, params):
        self.send_request(method, params)
        return self.recv_response()

    def send_request(self, method, params):
        self.id += 1
        self.send_json(
            dict(
                id=self.id,
                jsonrpc="2.0",
                method=method,
                params=params,
            )
        )

    def send_json(self, json_data):
        data = json.dumps(json_data)
        bindata = data.encode("utf-8")
        header = f"Content-Length: {len(bindata)}\r\n\r\n"

        self.stdin.write(header + data)
        self.stdin.flush()

    def recv_response(self):
        json_data = self.recv_json()
        assert json_data["id"] == self.id
        assert "error" not in json_data
        return json_data["result"]

    def send_notif(self, method, params):
        self.send_json(
            dict(
                jsonrpc="2.0",
                method=method,
                params=params,
            )
        )

    def expect_notif(self, method):
        json_data = self.recv_json()
        assert json_data["method"] == method
        assert "error" not in json_data
        return json_data["params"]

    def recv_json(self):
        header = self.stdout.readline()
        content_len_header = "Content-Length: "
        assert header.startswith(content_len_header)
        assert header.endswith("\r\n")
        data_len = int(header[len(content_len_header) : -2])

        # Expect end of header
        assert self.stdout.read(2) == "\r\n"

        data = self.stdout.read(data_len)
        return json.loads(data)

    def initialize(self, project):
        return self.request(
            "initialize",
            dict(
                processId=os.getpid(),
                rootUri="file://" + project,
                capabilities={},
            ),
        )

    def open_file(self, file):
        with open(file) as f:
            file_content = f.read()

        self.send_notif(
            "textDocument/didOpen",
            dict(
                textDocument=dict(
                    uri="file://" + os.path.realpath(file),
                    languageId="cpp",
                    version=1,
                    text=file_content,
                )
            ),
        )

    def show_diagnostics(self):
        diagnostics_response = self.expect_notif("textDocument/publishDiagnostics")
        uri = diagnostics_response["uri"]
        file_prefix = "file://"
        assert uri.startswith(file_prefix)
        file = uri[len(file_prefix) :]
        diagnostics = diagnostics_response["diagnostics"]
        for diagnostic in diagnostics:
            LSPClient.show_diagnostic(file, diagnostic)

        return len(diagnostics) != 0

    def show_diagnostic(file, diagnostic):
        range = diagnostic["range"]
        start = range["start"]
        line = start["line"]
        message = diagnostic["message"]
        print(f"{file}:{line}:{message}")

    def send_shutdown(self):
        self.send_request("shutdown", None)

    def recv_shutdown_send_exit(self):
        self.send_notif("exit", None)

    def wait(self):
        assert self.child.wait() == 0


def shutdown_all(clients):
    # Shutdown the clients in parallel. This drastically speeds up cleanup time.
    for client in clients:
        client.send_shutdown()

    for client in clients:
        client.recv_shutdown_send_exit()

    for client in clients:
        client.wait()


def get_clients(client_count, compile_commands_dir, total_jobs, verbose):
    # Distribute jobs evenly.
    # Clients near the front get more jobs, but they also may get more files.
    job_count = [0] * client_count
    for i in range(total_jobs):
        job_count[i % client_count] += 1
    return [LSPClient(compile_commands_dir=compile_commands_dir, jobs=jobs, verbose=verbose) for jobs in job_count]


def get_diagnostics(files, *, client_count, compile_commands_dir, total_jobs, verbose):
    if client_count > total_jobs:
        print(f"Client count {client_count} is greater than total jobs {total_jobs}. Forcing the client count to {total_jobs}.", file=sys.stderr)
        client_count = total_jobs

    project = os.getcwd()

    client_count = min(client_count, len(files))
    clients = get_clients(client_count, compile_commands_dir, total_jobs, verbose)
    for client in clients:
        client.initialize(project)

    # Similar to distributing jobs, we distribute files evenly.
    file_counts = [0] * client_count
    for i, file in enumerate(files):
        client_idx = i % client_count
        clients[client_idx].open_file(file)
        file_counts[client_idx] += 1

    any_diagnostic = 0
    for file_count, client in zip(file_counts, clients):
        for _ in range(file_count):
            any_diagnostic |= client.show_diagnostics()

    shutdown_all(clients)
    return any_diagnostic


def main():
    parser = argparse.ArgumentParser(
        prog="get-clangd-diagnostics.py",
        description="Scan project for any clangd diagnostics (including warnings) and outputs them.",
    )
    parser.add_argument("files", nargs="+", help="Files to scan")
    parser.add_argument(
        "--instances",
        type=int,
        default=4,
        help="Number of clangd instances to spawn in parallel. Defaults to 4.",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        type=int,
        help="Number of total jobs across all servers. Defaults to the CPU count.",
    )
    parser.add_argument(
        "-p",
        "--compile-commands-dir",
        help="Directory containining compile_commands.json",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Show clangd debug output",
    )

    args = parser.parse_args(sys.argv[1:])
    jobs = args.jobs if args.jobs is not None else multiprocessing.cpu_count()
    return get_diagnostics(
        args.files,
        client_count=args.instances,
        compile_commands_dir=args.compile_commands_dir,
        total_jobs=jobs,
        verbose=args.verbose,
    )


sys.exit(main())



from __future__ import annotations

import datetime
from decimal import Decimal
from uuid import UUID

import numpy as np
import pandas as pd
import pytz
from type_aliases import ConnDB


def test_bool(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.isStudent;")
    assert result.has_next()
    assert result.get_next() == [True]
    assert not result.has_next()
    result.close()


def test_int(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.age;")
    assert result.has_next()
    assert result.get_next() == [35]
    assert not result.has_next()
    result.close()


def test_int8(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) -[r:studyAt]-> (b:organisation) WHERE r.length = 5 RETURN r.level;")
    assert result.has_next()
    assert result.get_next() == [5]
    assert not result.has_next()
    result.close()


def test_uint8(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) -[r:studyAt]-> (b:organisation) WHERE r.length = 5 RETURN r.ulevel;")
    assert result.has_next()
    assert result.get_next() == [250]
    assert not result.has_next()
    result.close()


def test_uint16(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) -[r:studyAt]-> (b:organisation) WHERE r.length = 5 RETURN r.ulength;")
    assert result.has_next()
    assert result.get_next() == [33768]
    assert not result.has_next()
    result.close()


def test_uint32(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) -[r:studyAt]-> (b:organisation) WHERE r.length = 5 RETURN r.temperature;")
    assert result.has_next()
    assert result.get_next() == [32800]
    assert not result.has_next()
    result.close()


def test_uint64(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) -[r:studyAt]-> (b:organisation) WHERE r.length = 5 RETURN r.code;")
    assert result.has_next()
    assert result.get_next() == [9223372036854775808]
    assert not result.has_next()
    result.close()


def test_int128(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) -[r:studyAt]-> (b:organisation) WHERE r.length = 5 RETURN r.hugedata;")
    assert result.has_next()
    assert result.get_next() == [1844674407370955161811111111]
    assert not result.has_next()
    result.close()


def test_serial(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:moviesSerial) WHERE a.ID = 2 RETURN a.ID;")
    assert result.has_next()
    assert result.get_next() == [2]
    assert not result.has_next()
    result.close()


def test_double(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.eyeSight;")
    assert result.has_next()
    assert result.get_next() == [5.0]
    assert not result.has_next()
    result.close()


def test_decimal(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    res = conn.execute(
        "UNWIND [1, 2, 3] AS A UNWIND [5.7, 8.3, 2.9] AS B WITH cast(CAST(A AS DECIMAL) * CAST(B AS DECIMAL) AS DECIMAL(18, 1)) AS PROD RETURN COLLECT(PROD) AS RES"
    )
    assert sorted(res.get_next()[0]) == sorted([
        Decimal("5.7"),
        Decimal("8.3"),
        Decimal("2.9"),
        Decimal("11.4"),
        Decimal("16.6"),
        Decimal("5.8"),
        Decimal("17.1"),
        Decimal("24.9"),
        Decimal("8.7"),
    ])
    res = conn.execute(
        "UNWIND [1, 2, 3] AS A UNWIND [5.7, 8.3, 2.9] AS B WITH CAST(CAST(A AS DECIMAL) * CAST(B AS DECIMAL) AS DECIMAL(4, 1)) AS PROD RETURN COLLECT(PROD) AS RES"
    )
    assert sorted(res.get_next()[0]) == sorted([
        Decimal("5.7"),
        Decimal("8.3"),
        Decimal("2.9"),
        Decimal("11.4"),
        Decimal("16.6"),
        Decimal("5.8"),
        Decimal("17.1"),
        Decimal("24.9"),
        Decimal("8.7"),
    ])


def test_string(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.fName;")
    assert result.has_next()
    assert result.get_next() == ["Alice"]
    assert not result.has_next()
    result.close()


def test_blob(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("RETURN BLOB('\\\\xAA\\\\xBB\\\\xCD\\\\x1A')")
    assert result.has_next()
    result_blob = result.get_next()[0]
    assert len(result_blob) == 4
    assert result_blob[0] == 0xAA
    assert result_blob[1] == 0xBB
    assert result_blob[2] == 0xCD
    assert result_blob[3] == 0x1A
    assert not result.has_next()
    result.close()


def test_uuid(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("RETURN UUID('A0EEBC99-9c0b-4ef8-bb6d-6bb9bd380a12')")
    assert result.has_next()
    assert result.get_next() == [UUID("a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a12")]
    assert not result.has_next()
    result.close()


def test_date(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.birthdate;")
    assert result.has_next()
    assert result.get_next() == [datetime.date(1900, 1, 1)]
    assert not result.has_next()
    result.close()


def test_timestamp(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.registerTime;")
    assert result.has_next()
    assert result.get_next() == [datetime.datetime(2011, 8, 20, 11, 25, 30)]
    assert not result.has_next()
    result.close()


def test_timestamp_tz(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:movies) WHERE a.length = 126 RETURN a.description.release_tz;")
    assert result.has_next()
    assert result.get_next() == [datetime.datetime(2011, 8, 20, 11, 25, 30, 123456, pytz.UTC)]
    assert not result.has_next()
    result.close()


def test_timestamp_ns(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:movies) WHERE a.length = 126 RETURN a.description.release_ns;")
    assert result.has_next()
    assert result.get_next() == [datetime.datetime(2011, 8, 20, 11, 25, 30, 123456)]
    assert not result.has_next()
    result.close()


def test_timestamp_ms(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:movies) WHERE a.length = 126 RETURN a.description.release_ms;")
    assert result.has_next()
    assert result.get_next() == [datetime.datetime(2011, 8, 20, 11, 25, 30, 123000)]
    assert not result.has_next()
    result.close()


def test_timestamp_s(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:movies) WHERE a.length = 126 RETURN a.description.release_sec;")
    assert result.has_next()
    assert result.get_next() == [datetime.datetime(2011, 8, 20, 11, 25, 30)]
    assert not result.has_next()
    result.close()


def test_interval(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.lastJobDuration;")
    assert result.has_next()
    assert result.get_next() == [datetime.timedelta(days=1082, seconds=46920)]
    assert not result.has_next()
    result.close()


def test_list(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.courseScoresPerTerm;")
    assert result.has_next()
    assert result.get_next() == [[[10, 8], [6, 7, 8]]]
    assert not result.has_next()
    result.close()


def test_map(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (m:movies) WHERE m.length = 2544 RETURN m.audience;")
    assert result.has_next()
    assert result.get_next() == [{"audience1": 33}]
    assert not result.has_next()
    result.close()


def test_union(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (m:movies) WHERE m.length = 2544 RETURN m.grade;")
    assert result.has_next()
    assert result.get_next() == [8.989]
    assert not result.has_next()
    result.close()


def test_node(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a")
    assert result.has_next()
    n = result.get_next()
    assert len(n) == 1
    n = n[0]
    assert n["_label"] == "person"
    assert n["ID"] == 0
    assert n["fName"] == "Alice"
    assert n["gender"] == 1
    assert n["isStudent"] is True
    assert n["eyeSight"] == 5.0
    assert n["birthdate"] == datetime.date(1900, 1, 1)
    assert n["registerTime"] == datetime.datetime(2011, 8, 20, 11, 25, 30)
    assert n["lastJobDuration"] == datetime.timedelta(days=1082, seconds=46920)
    assert n["courseScoresPerTerm"] == [[10, 8], [6, 7, 8]]
    assert n["usedNames"] == ["Aida"]
    assert not result.has_next()
    result.close()


def test_rel(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (p:person)-[r:workAt]->(o:organisation) WHERE p.ID = 5 RETURN p, r, o")
    assert result.has_next()
    n = result.get_next()
    assert len(n) == 3
    p = n[0]
    r = n[1]
    o = n[2]
    assert p["_label"] == "person"
    assert p["ID"] == 5
    assert o["_label"] == "organisation"
    assert o["ID"] == 6
    assert r["year"] == 2010
    assert r["_src"] == p["_id"]
    assert r["_dst"] == o["_id"]
    assert r["_label"] == "workAt"
    assert not result.has_next()
    result.close()


def test_struct(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute('MATCH (m:movies) WHERE m.name="Roma" RETURN m.description')
    assert result.has_next()
    n = result.get_next()
    assert len(n) == 1
    description = n[0]
    print(description)
    assert description["rating"] == 1223
    assert description["views"] == 10003
    assert description["release"] == datetime.datetime(2011, 2, 11, 16, 44, 22)
    assert description["release_ns"] == datetime.datetime(2011, 2, 11, 16, 44, 22, 123456)
    assert description["release_ms"] == datetime.datetime(2011, 2, 11, 16, 44, 22, 123000)
    assert description["release_sec"] == datetime.datetime(2011, 2, 11, 16, 44, 22)
    assert description["release_tz"] == datetime.datetime(2011, 2, 11, 16, 44, 22, 123456, pytz.UTC)
    assert description["film"] == datetime.date(2013, 2, 22)
    assert description["stars"] == 100
    assert description["u8"] == 1
    assert description["u16"] == 15
    assert description["u32"] == 200
    assert description["u64"] == 4
    assert description["hugedata"] == -15
    assert not result.has_next()
    result.close()


def test_recursive_rel(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    result = conn.execute("MATCH (a:person)-[e:studyAt*1..1]->(b:organisation) WHERE a.fName = 'Alice' RETURN e;")
    assert result.has_next()
    n = result.get_next()
    assert len(n) == 1
    e = n[0]
    assert "_nodes" in e
    assert "_rels" in e
    assert len(e["_nodes"]) == 0
    assert len(e["_rels"]) == 1
    rel = e["_rels"][0]
    excepted_rel = {
        "_id": {"offset": 0, "table": 5},
        "_src": {"offset": 0, "table": 0},
        "_dst": {"offset": 0, "table": 1},
        "_label": "studyAt",
        "year": 2021,
        "places": ["wwAewsdndweusd", "wek"],
        "length": 5,
        "level": 5,
        "code": 9223372036854775808,
        "temperature": 32800,
        "ulength": 33768,
        "ulevel": 250,
        "hugedata": 1844674407370955161811111111,
    }
    assert rel == excepted_rel
    assert not result.has_next()
    result.close()


def test_large_array(conn_db_readwrite: ConnDB) -> None:
    conn, _ = conn_db_readwrite

    data = []
    for i in range(1000):
        data.append({"id": i, "embedding": np.random.rand(1670).tolist()})

    df = pd.DataFrame(data)
    conn.execute("CREATE NODE TABLE _User(id INT64, embedding DOUBLE[1670], PRIMARY KEY (id))")
    conn.execute("COPY _User FROM df")
    db_df = conn.execute("MATCH (u:_User) RETURN u.id as id, u.embedding as embedding ORDER BY u.id").get_as_df()
    sorted_df = df.sort_values(by="id").reset_index(drop=True)
    sorted_db_df = db_df.sort_values(by="id").reset_index(drop=True)
    assert sorted_df.equals(sorted_db_df)



import argparse
import subprocess
import os
import shutil


def create_worktree(path, commit, repo_root):
    remove_worktree(path, repo_root)
    run_command(f"git worktree add {path} {commit}", cwd=repo_root)


def remove_worktree(path, repo_root):
    if os.path.exists(path):
        run_command(f"git worktree remove --force {path}", cwd=repo_root)


def check_for_extension_build(makefile):
    with open(makefile, "r") as f:
        return any(line.strip() == "extension-build:" for line in f)


# Duplicates code from benchmark/version.py.
# Should be fine since the footprint is small, but any further extensions
# to this tool should rework this if the footprint for duplicated code gets
# bigger.
def get_version(kuzu_root):
    cmake_file = os.path.join(kuzu_root, "CMakeLists.txt")
    with open(cmake_file) as f:
        for line in f:
            if line.startswith("project(Kuzu VERSION"):
                return line.split(" ")[2].strip()
    return "0"


def run_command(cmd, cwd=None, capture_output=False):
    print(f"> Running: {cmd} (cwd={cwd})")

    # We redirect stdin to devnull in an attempt
    # to stop the process from interfering with the terminal's input buffer.
    # This needs some work. After running the script I found that my buffer
    # was filled with a sequence like 8;1R8;1R8;1R8;1R8;... This doesn't seem
    # to affect the script but is annoying.
    result = subprocess.run(
        cmd,
        cwd=cwd,
        text=True,
        capture_output=capture_output,
        check=True,
        shell=True,
        stdin=subprocess.DEVNULL,
    )
    if capture_output:
        return result.stdout.strip()
    else:
        print(result.stdout or "")
        return None


def main():
    parser = argparse.ArgumentParser(
        description="Export DBs from dataset-dir to output-dir using base-commit and test in test-commit"
    )
    parser.add_argument(
        "--base-commit", required=True, help="Git commit to export databases from"
    )
    parser.add_argument(
        "--test-commit", required=True, help="Git commit to test against"
    )
    parser.add_argument(
        "--dataset-dir", required=True, help="Path to the dataset directory"
    )
    parser.add_argument(
        "--output-dir", required=True, help="Path to output the exported databases"
    )
    parser.add_argument(
        "--cleanup",
        dest="cleanup",
        action="store_true",
        help="Delete exported DBs after test",
    )
    parser.add_argument(
        "--no-cleanup",
        dest="cleanup",
        action="store_false",
        help="Do not delete exported DBs after test",
    )
    parser.set_defaults(cleanup=True)
    args = parser.parse_args()

    base_commit = args.base_commit
    test_commit = args.test_commit
    dataset_dir = args.dataset_dir
    output_dir = args.output_dir
    cleanup = args.cleanup

    script_dir = os.path.dirname(os.path.realpath(__file__))
    kuzu_root = os.path.abspath(os.path.join(script_dir, ".."))
    base_worktree = os.path.join(kuzu_root, ".worktree-base")
    test_worktree = os.path.join(kuzu_root, ".worktree-test")
    create_worktree(base_worktree, base_commit, kuzu_root)
    create_worktree(test_worktree, test_commit, kuzu_root)
    export_path = None
    try:
        version = get_version(base_worktree)
        if version == "0":
            raise Exception("Failed to determine version. Aborting.")
        export_path = os.path.abspath(os.path.join(output_dir, version))

        if not os.path.exists(export_path + os.sep):
            # Some datasets, like tinysnb_json, have a dependency on the JSON
            # extension in their copy.cypher files. Therefore, we must build with
            # JSON support to ensure the export works correctly.

            # Older makefiles did not have the command specified under else
            if check_for_extension_build(
                os.path.abspath(os.path.join(base_worktree, "Makefile"))
            ):
                run_command(
                    "make extension-build EXTENSION_LIST=json", cwd=base_worktree
                )
            else:
                run_command(
                    "make extension-test-build EXTENSION_LIST=json", cwd=base_worktree
                )
            inprogress_path = f"{export_path}_inprogress" + os.sep
            export_script_path = os.path.join(kuzu_root, "scripts", "export-dbs.py")
            exec_path = os.path.join(
                base_worktree, "build", "relwithdebinfo", "tools", "shell", "kuzu"
            )
            run_command(
                f"""python3 {export_script_path} \
                --executable {exec_path} \
                --dataset-dir {dataset_dir} \
                --output-dir {inprogress_path}""",
                cwd=kuzu_root,
            )
            os.rename(inprogress_path, export_path + os.sep)

        # appends / so that datasets can be found correctly
        export_path += os.sep
        os.environ["E2E_IMPORT_DB_DIR"] = export_path
        run_command("make test", cwd=test_worktree)
        return 0
    finally:
        if cleanup and export_path and os.path.exists(export_path):
            print(f"Cleaning up export directory: {export_path}")
            shutil.rmtree(export_path)
        else:
            print(f"Skipping cleaning up export directory: {export_path}")

        print("Removing worktrees")
        remove_worktree(base_worktree, kuzu_root)
        remove_worktree(test_worktree, kuzu_root)


if __name__ == "__main__":
    main()



from datetime import date, datetime, timedelta

TINY_SNB_PERSONS_GROUND_TRUTH = {
    0: {
        "ID": 0,
        "fName": "Alice",
        "gender": 1,
        "isStudent": True,
        "isWorker": False,
        "age": 35,
        "eyeSight": 5.0,
        "birthdate": date(1900, 1, 1),
        "registerTime": datetime(2011, 8, 20, 11, 25, 30),
        "lastJobDuration": timedelta(days=1082, seconds=46920),
        "workedHours": [10, 5],
        "usedNames": ["Aida"],
        "courseScoresPerTerm": [[10, 8], [6, 7, 8]],
        "grades": [96, 54, 86, 92],
        "height": 1.731,
        "u": "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a11",
        "_LABEL": "person",
        "_ID": {"offset": 0, "table": 0},
    },
    2: {
        "ID": 2,
        "fName": "Bob",
        "gender": 2,
        "isStudent": True,
        "isWorker": False,
        "age": 30,
        "eyeSight": 5.1,
        "birthdate": date(1900, 1, 1),
        "registerTime": datetime(2008, 11, 3, 15, 25, 30, 526),
        "lastJobDuration": timedelta(days=3750, seconds=46800, microseconds=24),
        "workedHours": [12, 8],
        "usedNames": ["Bobby"],
        "courseScoresPerTerm": [[8, 9], [9, 10]],
        "grades": [98, 42, 93, 88],
        "height": 0.99,
        "u": "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a12",
        "_LABEL": "person",
        "_ID": {"offset": 1, "table": 0},
    },
    3: {
        "ID": 3,
        "fName": "Carol",
        "gender": 1,
        "isStudent": False,
        "isWorker": True,
        "age": 45,
        "eyeSight": 5.0,
        "birthdate": date(1940, 6, 22),
        "registerTime": datetime(1911, 8, 20, 2, 32, 21),
        "lastJobDuration": timedelta(days=2, seconds=1451),
        "workedHours": [4, 5],
        "usedNames": ["Carmen", "Fred"],
        "courseScoresPerTerm": [[8, 10]],
        "grades": [91, 75, 21, 95],
        "height": 1.00,
        "u": "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a13",
        "_LABEL": "person",
        "_ID": {"offset": 2, "table": 0},
    },
    5: {
        "ID": 5,
        "fName": "Dan",
        "gender": 2,
        "isStudent": False,
        "isWorker": True,
        "age": 20,
        "eyeSight": 4.8,
        "birthdate": date(1950, 7, 23),
        "registerTime": datetime(2031, 11, 30, 12, 25, 30),
        "lastJobDuration": timedelta(days=3750, seconds=46800, microseconds=24),
        "workedHours": [1, 9],
        "usedNames": ["Wolfeschlegelstein", "Daniel"],
        "courseScoresPerTerm": [[7, 4], [8, 8], [9]],
        "grades": [76, 88, 99, 89],
        "height": 1.30,
        "u": "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a14",
        "_LABEL": "person",
        "_ID": {"offset": 3, "table": 0},
    },
    7: {
        "ID": 7,
        "fName": "Elizabeth",
        "gender": 1,
        "isStudent": False,
        "isWorker": True,
        "age": 20,
        "eyeSight": 4.7,
        "birthdate": date(1980, 10, 26),
        "registerTime": datetime(1976, 12, 23, 11, 21, 42),
        "lastJobDuration": timedelta(days=2, seconds=1451),
        "workedHours": [2],
        "usedNames": ["Ein"],
        "courseScoresPerTerm": [[6], [7], [8]],
        "grades": [96, 59, 65, 88],
        "height": 1.463,
        "u": "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a15",
        "_LABEL": "person",
        "_ID": {"offset": 4, "table": 0},
    },
    8: {
        "ID": 8,
        "fName": "Farooq",
        "gender": 2,
        "isStudent": True,
        "isWorker": False,
        "age": 25,
        "eyeSight": 4.5,
        "birthdate": date(1980, 10, 26),
        "registerTime": datetime(1972, 7, 31, 13, 22, 30, 678559),
        "lastJobDuration": timedelta(seconds=1080, microseconds=24000),
        "workedHours": [3, 4, 5, 6, 7],
        "usedNames": ["Fesdwe"],
        "courseScoresPerTerm": [[8]],
        "grades": [80, 78, 34, 83],
        "height": 1.51,
        "u": "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a16",
        "_LABEL": "person",
        "_ID": {"offset": 5, "table": 0},
    },
    9: {
        "ID": 9,
        "fName": "Greg",
        "gender": 2,
        "isStudent": False,
        "isWorker": False,
        "age": 40,
        "eyeSight": 4.9,
        "birthdate": date(1980, 10, 26),
        "registerTime": datetime(1976, 12, 23, 4, 41, 42),
        "lastJobDuration": timedelta(days=3750, seconds=46800, microseconds=24),
        "workedHours": [1],
        "usedNames": ["Grad"],
        "courseScoresPerTerm": [[10]],
        "grades": [43, 83, 67, 43],
        "height": 1.6,
        "u": "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a17",
        "_LABEL": "person",
        "_ID": {"offset": 6, "table": 0},
    },
    10: {
        "ID": 10,
        "fName": "Hubert Blaine Wolfeschlegelsteinhausenbergerdorff",
        "gender": 2,
        "isStudent": False,
        "isWorker": True,
        "age": 83,
        "eyeSight": 4.9,
        "birthdate": date(1990, 11, 27),
        "registerTime": datetime(2023, 2, 21, 13, 25, 30),
        "lastJobDuration": timedelta(days=1082, seconds=46920),
        "workedHours": [10, 11, 12, 3, 4, 5, 6, 7],
        "usedNames": ["Ad", "De", "Hi", "Kye", "Orlan"],
        "courseScoresPerTerm": [[7], [10], [6, 7]],
        "grades": [77, 64, 100, 54],
        "height": 1.323,
        "u": "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a18",
        "_LABEL": "person",
        "_ID": {"offset": 7, "table": 0},
    },
}

TINY_SNB_ORGANISATIONS_GROUND_TRUTH = {
    1: {
        "ID": 1,
        "name": "ABFsUni",
        "orgCode": 325,
        "mark": 3.7,
        "score": -2,
        "history": "10 years 5 months 13 hours 24 us",
        "licenseValidInterval": timedelta(days=1085),
        "rating": 1.0,
        "state": {"revenue": 138, "location": ["toronto", "montr,eal"], "stock": {"price": [96, 56], "volume": 1000}},
        "info": 3.12,
        "_LABEL": "organisation",
        "_ID": {"offset": 0, "table": 1},
    },
    4: {
        "ID": 4,
        "name": "CsWork",
        "orgCode": 934,
        "mark": 4.1,
        "score": -100,
        "history": "2 years 4 days 10 hours",
        "licenseValidInterval": timedelta(days=9414),
        "rating": 0.78,
        "state": {
            "revenue": 152,
            "location": ['"vanco,uver north area"'],
            "stock": {"price": [15, 78, 671], "volume": 432},
        },
        "info": "abcd",
        "_LABEL": "organisation",
        "_ID": {"offset": 1, "table": 1},
    },
    6: {
        "ID": 6,
        "name": "DEsWork",
        "orgCode": 824,
        "mark": 4.1,
        "score": 7,
        "history": "2 years 4 hours 22 us 34 minutes",
        "licenseValidInterval": timedelta(days=3, seconds=36000, microseconds=100000),
        "rating": 0.52,
        "state": {
            "revenue": 558,
            "location": ["'very long city name'", "'new york'"],
            "stock": {"price": [22], "volume": 99},
        },
        "info": date(2023, 12, 15),
        "_LABEL": "organisation",
        "_ID": {"offset": 2, "table": 1},
    },
}

TINY_SNB_KNOWS_GROUND_TRUTH = {
    0: [2, 3, 5],
    2: [0, 3, 5],
    3: [0, 2, 5],
    5: [0, 2, 3],
    7: [8, 9],
}

TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH = {
    (0, 2): {
        "date": date(2021, 6, 30),
        "meetTime": datetime(1986, 10, 21, 21, 8, 31, 521000),
        "validInterval": timedelta(days=3750, seconds=46800, microseconds=24),
        "comments": ["rnme", "m8sihsdnf2990nfiwf"],
        "summary": {
            "locations": ["'toronto'", "'waterloo'"],
            "transfer": {"day": date(2021, 1, 2), "amount": [100, 200]},
        },
        "notes": 1,
        "someMap": {"a": "b"},
    },
    (0, 3): {
        "date": date(2021, 6, 30),
        "meetTime": datetime(1946, 8, 25, 19, 7, 22),
        "validInterval": timedelta(days=7232),
        "comments": ["njnojppo9u0jkmf", "fjiojioh9h9h89hph"],
        "summary": {"locations": None, "transfer": None},
        "notes": date(2020, 10, 10),
        "someMap": {"c": "d", "e": "f", "1": "2"},
    },
    (0, 5): {
        "date": date(2021, 6, 30),
        "meetTime": datetime(2012, 12, 11, 20, 7, 22),
        "validInterval": timedelta(days=10),
        "comments": ["ioji232", "jifhe8w99u43434"],
        "summary": {"locations": ["'shanghai'"], "transfer": {"day": date(1990, 9, 10), "amount": [10]}},
        "notes": "nice weather",
        "someMap": None,
    },
    (2, 0): {
        "date": date(2021, 6, 30),
        "meetTime": datetime(1946, 8, 25, 19, 7, 22),
        "validInterval": timedelta(days=3750, seconds=46800, microseconds=24),
        "comments": ["2huh9y89fsfw23", "23nsihufhw723"],
        "summary": {"locations": ["'paris'"], "transfer": {"day": date(2000, 1, 1), "amount": [20, 5000]}},
        "notes": 4,
        "someMap": None,
    },
    (2, 3): {
        "date": date(1950, 5, 14),
        "meetTime": datetime(1946, 8, 25, 19, 7, 22),
        "validInterval": timedelta(seconds=1380),
        "comments": ["fwehu9h9832wewew", "23u9h989sdfsss"],
        "summary": {"locations": ["'paris'"], "transfer": {"day": date(2011, 5, 1), "amount": [2000, 5340]}},
        "notes": "cool stuff found",
        "someMap": None,
    },
    (2, 5): {
        "date": date(1950, 5, 14),
        "meetTime": datetime(2012, 12, 11, 20, 7, 22),
        "validInterval": timedelta(days=7232),
        "comments": ["fwh9y81232uisuiehuf", "ewnuihxy8dyf232"],
        "summary": {"locations": ["'vancouver'"], "transfer": {"day": date(2020, 1, 1), "amount": [120, 50]}},
        "notes": "matthew perry",
        "someMap": None,
    },
    (3, 0): {
        "date": date(2021, 6, 30),
        "meetTime": datetime(2002, 7, 31, 11, 42, 53, 123420),
        "validInterval": timedelta(days=41, seconds=21600),
        "comments": ["fnioh8323aeweae34d", "osd89e2ejshuih12"],
        "summary": {
            "locations": ["'london'", "'toronto'"],
            "transfer": {"day": date(2012, 11, 21), "amount": [223, 5230]},
        },
        "notes": 10,
        "someMap": None,
    },
    (3, 2): {
        "date": date(1950, 5, 14),
        "meetTime": datetime(2007, 2, 12, 12, 11, 42, 123000),
        "validInterval": timedelta(seconds=1680, microseconds=30000),
        "comments": ["fwh983-sdjisdfji", "ioh89y32r2huir"],
        "summary": {
            "locations": ["'paris'", "'beijing'"],
            "transfer": {"day": date(2011, 3, 11), "amount": [2323, 50]},
        },
        "notes": 1,
        "someMap": None,
    },
    (3, 5): {
        "date": date(2000, 1, 1),
        "meetTime": datetime(1998, 10, 2, 13, 9, 22, 423000),
        "validInterval": timedelta(microseconds=300000),
        "comments": ["psh989823oaaioe", "nuiuah1nosndfisf"],
        "summary": {"locations": [], "transfer": {"day": date(1980, 11, 21), "amount": [20, 5]}},
        "notes": 2,
        "someMap": None,
    },
    (5, 0): {
        "date": date(2021, 6, 30),
        "meetTime": datetime(1936, 11, 2, 11, 2, 1),
        "validInterval": timedelta(microseconds=480),
        "comments": ["fwewe"],
        "summary": {
            "locations": ["'shanghai'", "'nanjing'"],
            "transfer": {"day": date(1998, 11, 12), "amount": [22, 53240]},
        },
        "notes": 15,
        "someMap": None,
    },
    (5, 2): {
        "date": date(1950, 5, 14),
        "meetTime": datetime(1982, 11, 11, 13, 12, 5, 123000),
        "validInterval": timedelta(seconds=1380),
        "comments": ["fewh9182912e3", "h9y8y89soidfsf", "nuhudf78w78efw", "hioshe0f9023sdsd"],
        "summary": {"locations": ["'paris'"], "transfer": {"day": date(2000, 1, 1), "amount": [20, 5000]}},
        "notes": "happy new year",
        "someMap": None,
    },
    (5, 3): {
        "date": date(2000, 1, 1),
        "meetTime": datetime(1999, 4, 21, 15, 12, 11, 420000),
        "validInterval": timedelta(days=2, microseconds=52000),
        "comments": ["23h9sdslnfowhu2932", "shuhf98922323sf"],
        "summary": {"locations": ["'paris'"], "transfer": {"day": date(2000, 1, 1), "amount": [20, 5000]}},
        "notes": 4,
        "someMap": None,
    },
    (7, 8): {
        "date": date(1905, 12, 12),
        "meetTime": datetime(2025, 1, 1, 11, 22, 33, 520000),
        "validInterval": timedelta(seconds=2878),
        "comments": ["ahu2333333333333", "12weeeeeeeeeeeeeeeeee"],
        "summary": {
            "locations": ["'toronto'", "'thisisalongcityname'"],
            "transfer": {"day": date(1930, 11, 22), "amount": [18, 323]},
        },
        "notes": 8,
        "someMap": None,
    },
    (7, 9): {
        "date": date(1905, 12, 12),
        "meetTime": datetime(2020, 3, 1, 12, 11, 41, 655200),
        "validInterval": timedelta(seconds=2878),
        "comments": ["peweeeeeeeeeeeeeeeee", "kowje9w0eweeeeeeeee"],
        "summary": {
            "locations": ["'waterloo'"],
            "transfer": {"day": date(2000, 1, 1), "amount": [1000, 5000]},
        },
        "notes": 10,
        "someMap": None,
    },
}

TINY_SNB_WORKS_AT_GROUND_TRUTH = {
    3: [4],
    5: [6],
    7: [6],
}

TINY_SNB_WORKS_AT_PROPERTIES_GROUND_TRUTH = {(3, 4): {"year": 2015}, (5, 6): {"year": 2010}, (7, 6): {"year": 2015}}

TENSOR_LIST_GROUND_TRUTH = {
    0: {
        "boolTensor": [True, False],
        "doubleTensor": [[0.1, 0.2], [0.3, 0.4]],
        "intTensor": [[[1, 2], [3, 4]], [[5, 6], [7, 8]]],
    },
    3: {
        "boolTensor": [True, False],
        "doubleTensor": [[0.1, 0.2], [0.3, 0.4]],
        "intTensor": [[[3, 4], [5, 6]], [[7, 8], [9, 10]]],
    },
    4: {
        "boolTensor": [False, True],
        "doubleTensor": [[0.4, 0.8], [0.7, 0.6]],
        "intTensor": [[[5, 6], [7, 8]], [[9, 10], [11, 12]]],
    },
    5: {
        "boolTensor": [True, True],
        "doubleTensor": [[0.4, 0.9], [0.5, 0.2]],
        "intTensor": [[[7, 8], [9, 10]], [[11, 12], [13, 14]]],
    },
    6: {
        "boolTensor": [False, True],
        "doubleTensor": [[0.2, 0.4], [0.5, 0.1]],
        "intTensor": [[[9, 10], [11, 12]], [[13, 14], [15, 16]]],
    },
    8: {
        "boolTensor": [False, True],
        "doubleTensor": [[0.6, 0.4], [0.6, 0.1]],
        "intTensor": [[[11, 12], [13, 14]], [[15, 16], [17, 18]]],
    },
}

PERSONLONGSTRING_GROUND_TRUTH = {
    "AAAAAAAAAAAAAAAAAAAA": {
        "name": "AAAAAAAAAAAAAAAAAAAA",
        "spouse": "Bob",
    },
    "Bob": {
        "name": "Bob",
        "spouse": "AAAAAAAAAAAAAAAAAAAA",
    },
}

PERSONLONGSTRING_KNOWS_GROUND_TRUTH = {
    "AAAAAAAAAAAAAAAAAAAA": ["Bob"],
}



from __future__ import annotations

import math
from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import Any

import ground_truth
import kuzu
import polars as pl
import pyarrow as pa
import pytz
from pandas import Timestamp
from type_aliases import ConnDB

_expected_dtypes = {
    # ------------------------------------------------
    # person
    # ------------------------------------------------
    "a.ID": {"arrow": pa.int64(), "pl": pl.Int64},
    "a.fName": {"arrow": pa.string(), "pl": pl.String},
    "a.gender": {"arrow": pa.int64(), "pl": pl.Int64},
    "a.isStudent": {"arrow": pa.bool_(), "pl": pl.Boolean},
    "a.isWorker": {"arrow": pa.bool_(), "pl": pl.Boolean},
    "a.age": {"arrow": pa.int64(), "pl": pl.Int64},
    "a.eyeSight": {"arrow": pa.float64(), "pl": pl.Float64},
    "a.birthdate": {"arrow": pa.date32(), "pl": pl.Date},
    "a.registerTime": {"arrow": pa.timestamp("us"), "pl": pl.Datetime("us")},
    "a.lastJobDuration": {"arrow": pa.duration("us"), "pl": pl.Duration("us")},
    "a.workedHours": {"arrow": pa.list_(pa.int64()), "pl": pl.List(pl.Int64)},
    "a.usedNames": {"arrow": pa.list_(pa.string()), "pl": pl.List(pl.String)},
    "a.courseScoresPerTerm": {"arrow": pa.list_(pa.list_(pa.int64())), "pl": pl.List(pl.List(pl.Int64))},
    "a.grades": {"arrow": pa.list_(pa.int64(), 4), "pl": pl.Array(pl.Int64, shape=(4,))},
    "a.height": {"arrow": pa.float32(), "pl": pl.Float32},
    "a.u": {"arrow": pa.string(), "pl": pl.String},
    # ------------------------------------------------
    # movies
    # ------------------------------------------------
    "a.length": {"arrow": pa.int32(), "pl": pl.Int32},
    "m.name": {"arrow": pa.string(), "pl": pl.String},
    "a.description": {
        "arrow": pa.struct([
            ("rating", pa.float64()),
            ("stars", pa.int8()),
            ("views", pa.int64()),
            ("release", pa.timestamp("us")),
            ("release_ns", pa.timestamp("ns")),
            ("release_ms", pa.timestamp("ms")),
            ("release_sec", pa.timestamp("s")),
            ("release_tz", pa.timestamp("us", tz="UTC")),
            ("film", pa.date32()),
            ("u8", pa.uint8()),
            ("u16", pa.uint16()),
            ("u32", pa.uint32()),
            ("u64", pa.uint64()),
            ("hugedata", pa.decimal128(38, 0)),
        ]),
        "pl": pl.Struct({
            "rating": pl.Float64,
            "stars": pl.Int8,
            "views": pl.Int64,
            "release": pl.Datetime(time_unit="us"),
            "release_ns": pl.Datetime(time_unit="ns"),
            "release_ms": pl.Datetime(time_unit="ms"),
            "release_sec": pl.Datetime(time_unit="ms"),
            "release_tz": pl.Datetime(time_unit="us", time_zone="UTC"),
            "film": pl.Date,
            "u8": pl.UInt8,
            "u16": pl.UInt16,
            "u32": pl.UInt32,
            "u64": pl.UInt64,
            "hugedata": pl.Decimal(precision=38, scale=0),
        }),
    },
    # ------------------------------------------------
    # miscellaneous
    # ------------------------------------------------
    "a.lbl": {"arrow": pa.string(), "pl": pl.String},
    "a.orgCode": {"arrow": pa.int64(), "pl": pl.Int64},
}


def get_result(query_result: kuzu.QueryResult, result_type: str, chunk_size: int | None) -> Any:
    sz = [] if (chunk_size is None or result_type == "pl") else [chunk_size]
    res = getattr(query_result, f"get_as_{result_type}")(*sz)
    if result_type == "arrow" and chunk_size:
        assert res[0].num_chunks == max((len(res) // chunk_size), 1)
    return res


def assert_column_equals(data: Any, col_name: str, return_type: str, expected_values: list[Any]) -> None:
    col = data[col_name]
    col_dtype = col.dtype if hasattr(col, "dtype") else col.type
    values = col.to_pylist() if return_type == "arrow" else col.to_list()

    assert len(col) == len(expected_values)
    assert values == expected_values, f"Unexpected values for {col_name} ({return_type!r})"
    assert col_dtype == _expected_dtypes[col_name][return_type], f"Unexpected dtype for {col_name} ({return_type!r})"


def assert_col_names(data: Any, expected_col_names: list[str]) -> None:
    col_names = [(c._name if hasattr(c, "_name") else c.name) for c in data]
    assert col_names == expected_col_names, f"Unexpected column names: {col_names!r}"


def test_to_arrow(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly

    def _test_person_table(_conn: kuzu.Connection, return_type: str, chunk_size: int | None = None) -> None:
        query = "MATCH (a:person) RETURN a.* ORDER BY a.ID"
        data = get_result(_conn.execute(query), return_type, chunk_size)
        assert len(data.columns) == 16

        assert_column_equals(
            data=data,
            col_name="a.ID",
            return_type=return_type,
            expected_values=[0, 2, 3, 5, 7, 8, 9, 10],
        )

        assert_column_equals(
            data=data,
            col_name="a.fName",
            return_type=return_type,
            expected_values=[
                "Alice",
                "Bob",
                "Carol",
                "Dan",
                "Elizabeth",
                "Farooq",
                "Greg",
                "Hubert Blaine Wolfeschlegelsteinhausenbergerdorff",
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.gender",
            return_type=return_type,
            expected_values=[1, 2, 1, 2, 1, 2, 2, 2],
        )

        assert_column_equals(
            data=data,
            col_name="a.isStudent",
            return_type=return_type,
            expected_values=[True, True, False, False, False, True, False, False],
        )

        assert_column_equals(
            data=data,
            col_name="a.isWorker",
            return_type=return_type,
            expected_values=[False, False, True, True, True, False, False, True],
        )

        assert_column_equals(
            data=data,
            col_name="a.age",
            return_type=return_type,
            expected_values=[35, 30, 45, 20, 20, 25, 40, 83],
        )

        assert_column_equals(
            data=data,
            col_name="a.eyeSight",
            return_type=return_type,
            expected_values=[5.0, 5.1, 5.0, 4.8, 4.7, 4.5, 4.9, 4.9],
        )

        assert_column_equals(
            data=data,
            col_name="a.birthdate",
            return_type=return_type,
            expected_values=[
                date(1900, 1, 1),
                date(1900, 1, 1),
                date(1940, 6, 22),
                date(1950, 7, 23),
                date(1980, 10, 26),
                date(1980, 10, 26),
                date(1980, 10, 26),
                date(1990, 11, 27),
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.registerTime",
            return_type=return_type,
            expected_values=[
                datetime(2011, 8, 20, 11, 25, 30),
                datetime(2008, 11, 3, 15, 25, 30, 526),
                datetime(1911, 8, 20, 2, 32, 21),
                datetime(2031, 11, 30, 12, 25, 30),
                datetime(1976, 12, 23, 11, 21, 42),
                datetime(1972, 7, 31, 13, 22, 30, 678559),
                datetime(1976, 12, 23, 4, 41, 42),
                datetime(2023, 2, 21, 13, 25, 30),
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.lastJobDuration",
            return_type=return_type,
            expected_values=[
                timedelta(days=1082, seconds=46920),
                timedelta(days=3750, seconds=46800, microseconds=24),
                timedelta(days=2, seconds=1451),
                timedelta(days=3750, seconds=46800, microseconds=24),
                timedelta(days=2, seconds=1451),
                timedelta(seconds=1080, microseconds=24000),
                timedelta(days=3750, seconds=46800, microseconds=24),
                timedelta(days=1082, seconds=46920),
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.workedHours",
            return_type=return_type,
            expected_values=[[10, 5], [12, 8], [4, 5], [1, 9], [2], [3, 4, 5, 6, 7], [1], [10, 11, 12, 3, 4, 5, 6, 7]],
        )

        assert_column_equals(
            data=data,
            col_name="a.usedNames",
            return_type=return_type,
            expected_values=[
                ["Aida"],
                ["Bobby"],
                ["Carmen", "Fred"],
                ["Wolfeschlegelstein", "Daniel"],
                ["Ein"],
                ["Fesdwe"],
                ["Grad"],
                ["Ad", "De", "Hi", "Kye", "Orlan"],
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.courseScoresPerTerm",
            return_type=return_type,
            expected_values=[
                [[10, 8], [6, 7, 8]],
                [[8, 9], [9, 10]],
                [[8, 10]],
                [[7, 4], [8, 8], [9]],
                [[6], [7], [8]],
                [[8]],
                [[10]],
                [[7], [10], [6, 7]],
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.grades",
            return_type=return_type,
            expected_values=[
                [96, 54, 86, 92],
                [98, 42, 93, 88],
                [91, 75, 21, 95],
                [76, 88, 99, 89],
                [96, 59, 65, 88],
                [80, 78, 34, 83],
                [43, 83, 67, 43],
                [77, 64, 100, 54],
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.u",
            return_type=return_type,
            expected_values=[
                "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a11",
                "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a12",
                "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a13",
                "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a14",
                "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a15",
                "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a16",
                "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a17",
                "a0eebc99-9c0b-4ef8-bb6d-6bb9bd380a18",
            ],
        )

    def _test_movies_table(_conn: kuzu.Connection, return_type: str, chunk_size: int | None = None) -> None:
        query = "MATCH (a:movies) RETURN a.length, a.description ORDER BY a.length"
        data = get_result(_conn.execute(query), return_type, chunk_size)

        assert_col_names(data, ["a.length", "a.description"])
        assert_column_equals(
            data=data,
            col_name="a.length",
            return_type=return_type,
            expected_values=[126, 298, 2544],
        )

        assert_column_equals(
            data=data,
            col_name="a.description",
            return_type=return_type,
            expected_values=[
                {
                    "rating": 5.3,
                    "stars": 2,
                    "views": 152,
                    "release": datetime(2011, 8, 20, 11, 25, 30),
                    "release_ns": datetime(2011, 8, 20, 11, 25, 30, 123456)
                    if return_type == "pl"
                    else Timestamp("2011-08-20 11:25:30.123456"),
                    "release_ms": datetime(2011, 8, 20, 11, 25, 30, 123000),
                    "release_sec": datetime(2011, 8, 20, 11, 25, 30),
                    "release_tz": datetime(2011, 8, 20, 11, 25, 30, 123456, pytz.UTC),
                    "film": date(2012, 5, 11),
                    "u8": 220,
                    "u16": 20,
                    "u32": 1,
                    "u64": 180,
                    "hugedata": Decimal(1844674407370955161811111111),
                },
                {
                    "rating": 1223.0,
                    "stars": 100,
                    "views": 10003,
                    "release": datetime(2011, 2, 11, 16, 44, 22),
                    "release_ns": datetime(2011, 2, 11, 16, 44, 22, 123456)
                    if return_type == "pl"
                    else Timestamp("2011-02-11 16:44:22.123456"),
                    "release_ms": datetime(2011, 2, 11, 16, 44, 22, 123000),
                    "release_sec": datetime(2011, 2, 11, 16, 44, 22),
                    "release_tz": datetime(2011, 2, 11, 16, 44, 22, 123456, pytz.UTC),
                    "film": date(2013, 2, 22),
                    "u8": 1,
                    "u16": 15,
                    "u32": 200,
                    "u64": 4,
                    "hugedata": Decimal(-15),
                },
                {
                    "rating": 7.0,
                    "stars": 10,
                    "views": 982,
                    "release": datetime(2018, 11, 13, 13, 33, 11),
                    "release_ns": datetime(2018, 11, 13, 13, 33, 11, 123456)
                    if return_type == "pl"
                    else Timestamp("2018-11-13 13:33:11.123456"),
                    "release_ms": datetime(2018, 11, 13, 13, 33, 11, 123000),
                    "release_sec": datetime(2018, 11, 13, 13, 33, 11),
                    "release_tz": datetime(2018, 11, 13, 13, 33, 11, 123456, pytz.UTC),
                    "film": date(2014, 9, 12),
                    "u8": 12,
                    "u16": 120,
                    "u32": 55,
                    "u64": 1,
                    "hugedata": Decimal(-1844674407370955161511),
                },
            ],
        )

    def _test_utf8_string(_conn: kuzu.Connection, return_type: str, chunk_size: int | None = None) -> None:
        query = "MATCH (m:movies) RETURN m.name"
        data = get_result(_conn.execute(query), return_type, chunk_size)

        assert_col_names(data, ["m.name"])
        assert_column_equals(
            data=data,
            col_name="m.name",
            return_type=return_type,
            expected_values=["Sóló cón tu párejâ", "The 😂😃🧘🏻‍♂️🌍🌦️🍞🚗 movie", "Roma"],
        )

    def _test_in_small_chunk_size(_conn: kuzu.Connection, return_type: str, chunk_size: int | None = None) -> None:
        query = "MATCH (a:person) RETURN a.age, a.fName ORDER BY a.ID"
        data = get_result(_conn.execute(query), return_type, chunk_size)

        assert_col_names(data, ["a.age", "a.fName"])
        assert_column_equals(
            data=data,
            col_name="a.age",
            return_type=return_type,
            expected_values=[35, 30, 45, 20, 20, 25, 40, 83],
        )

        assert_column_equals(
            data=data,
            col_name="a.fName",
            return_type=return_type,
            expected_values=[
                "Alice",
                "Bob",
                "Carol",
                "Dan",
                "Elizabeth",
                "Farooq",
                "Greg",
                "Hubert Blaine Wolfeschlegelsteinhausenbergerdorff",
            ],
        )

    def _test_with_nulls(_conn: kuzu.Connection, return_type: str, chunk_size: int | None = None) -> None:
        query = "MATCH (a:person:organisation) RETURN label(a) AS `a.lbl`, a.fName, a.orgCode ORDER BY a.ID"
        data = get_result(_conn.execute(query), return_type, chunk_size)

        assert_col_names(data, ["a.lbl", "a.fName", "a.orgCode"])
        assert_column_equals(
            data=data,
            col_name="a.lbl",
            return_type=return_type,
            expected_values=[
                "person",
                "organisation",
                "person",
                "person",
                "organisation",
                "person",
                "organisation",
                "person",
                "person",
                "person",
                "person",
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.fName",
            return_type=return_type,
            expected_values=[
                "Alice",
                None,
                "Bob",
                "Carol",
                None,
                "Dan",
                None,
                "Elizabeth",
                "Farooq",
                "Greg",
                "Hubert Blaine Wolfeschlegelsteinhausenbergerdorff",
            ],
        )

        assert_column_equals(
            data=data,
            col_name="a.orgCode",
            return_type=return_type,
            expected_values=[None, 325, None, None, 934, None, 824, None, None, None, None],
        )

    _test_person_table(conn, "arrow", 9)
    _test_person_table(conn, "pl")
    _test_movies_table(conn, "arrow", 8)
    _test_movies_table(conn, "pl")
    _test_utf8_string(conn, "arrow", 3)
    _test_utf8_string(conn, "pl")
    _test_in_small_chunk_size(conn, "arrow", 4)
    _test_in_small_chunk_size(conn, "pl", 4)
    _test_with_nulls(conn, "arrow", 12)
    _test_with_nulls(conn, "pl")


def test_to_arrow_map(conn_db_readonly: ConnDB) -> None:
    conn = conn_db_readonly[0]
    results = (
        conn.execute("RETURN map([1, 2, 3], [{a: 1, b: 2, c: '3'}, {a: 2, b: 3, c: '4'}, {a: 3, b: 4, c: '5'}])")
        .get_as_arrow(8)[0]
        .to_pylist()
    )
    assert results == [
        [(1, {"a": 1, "b": 2, "c": "3"}), (2, {"a": 2, "b": 3, "c": "4"}), (3, {"a": 3, "b": 4, "c": "5"})]
    ]


def test_to_arrow_array(conn_db_readwrite: ConnDB) -> None:
    conn = conn_db_readwrite[0]
    conn.execute("CREATE NODE TABLE school (id int64 primary key, prop1 int64[3])")
    conn.execute("CREATE (p:school {id: 5})")
    conn.execute("CREATE (p:school {id: 8, prop1: [2,3,1]})")
    conn.execute("CREATE (p:school {id: 18, prop1: null})")
    results = conn.execute("match (p:school) return p.*").get_as_arrow(8)
    assert results[0].to_pylist() == [5, 8, 18]
    assert results[1].to_pylist() == [None, [2, 3, 1], None]


def test_to_arrow_complex(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly

    def _test_node_helper(srcStruct, dstStruct):
        assert set(srcStruct.keys()) == set(dstStruct.keys())
        for key in srcStruct:
            if key == "_ID":  # there isn't any guarantee on the value of _ID, so ignore it
                continue
            if type(srcStruct[key]) is float:
                assert math.fabs(srcStruct[key] - dstStruct[key]) < 1e-5
            else:
                assert srcStruct[key] == dstStruct[key]

    def _test_node(_conn: kuzu.Connection) -> None:
        query = "MATCH (p:person) RETURN p ORDER BY p.ID"
        query_result = _conn.execute(query)
        arrow_tbl = query_result.get_as_arrow()
        p_col = arrow_tbl.column(0)

        for a, b in zip(
            p_col.to_pylist(),
            [
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[0],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[2],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[3],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[5],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[7],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[8],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[9],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[10],
            ],
        ):
            _test_node_helper(a, b)

    def _test_node_rel(_conn: kuzu.Connection) -> None:
        query = "MATCH (a:person)-[e:workAt]->(b:organisation) RETURN a, e, b ORDER BY a.ID, b.ID"
        query_result = _conn.execute(query)
        arrow_tbl = query_result.get_as_arrow(3)
        assert arrow_tbl.num_columns == 3
        a_col = arrow_tbl.column(0)
        assert len(a_col) == 3
        e_col = arrow_tbl.column(1)
        assert len(e_col) == 3
        b_col = arrow_tbl.column(2)
        assert len(b_col) == 3
        for a, b in zip(
            a_col.to_pylist(),
            [
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[3],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[5],
                ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[7],
            ],
        ):
            _test_node_helper(a, b)
        for a, b in zip(
            e_col.to_pylist(),
            [
                {
                    "_SRC": {"offset": 2, "table": 0},
                    "_DST": {"offset": 1, "table": 1},
                    "_ID": {"offset": 0, "table": 5},
                    "_LABEL": "workAt",
                    "grading": [3.8, 2.5],
                    "rating": 8.2,
                    "year": 2015,
                },
                {
                    "_SRC": {"offset": 3, "table": 0},
                    "_DST": {"offset": 2, "table": 1},
                    "_ID": {"offset": 1, "table": 5},
                    "_LABEL": "workAt",
                    "grading": [2.1, 4.4],
                    "rating": 7.6,
                    "year": 2010,
                },
                {
                    "_SRC": {"offset": 4, "table": 0},
                    "_DST": {"offset": 2, "table": 1},
                    "_ID": {"offset": 2, "table": 5},
                    "_LABEL": "workAt",
                    "grading": [9.2, 3.1],
                    "rating": 9.2,
                    "year": 2015,
                },
            ],
        ):
            _test_node_helper(a, b)

        for a, b in zip(
            b_col.to_pylist(),
            [
                ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[4],
                ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[6],
                ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[6],
            ],
        ):
            _test_node_helper(a, b)

    def _test_marries_table(_conn: kuzu.Connection) -> None:
        query = "MATCH (a:person)-[e:marries]->(b:person) RETURN e.* ORDER BY a.ID, b.ID"
        arrow_tbl = _conn.execute(query).get_as_arrow(0)
        assert arrow_tbl.num_columns == 3

        used_addr_col = arrow_tbl.column(0)
        assert used_addr_col.type == pa.list_(pa.string())
        assert len(used_addr_col) == 3
        assert used_addr_col.to_pylist() == [["toronto"], None, []]

        addr_col = arrow_tbl.column(1)
        assert addr_col.type == pa.list_(pa.int16(), 2)
        assert len(addr_col) == 3
        assert addr_col.to_pylist() == [[4, 5], [2, 5], [3, 9]]

        note_col = arrow_tbl.column(2)
        assert note_col.type == pa.string()
        assert len(note_col) == 3
        assert note_col.to_pylist() == [None, "long long long string", "short str"]

    def _test_recursive_rel(_conn: kuzu.Connection) -> None:
        query = "MATCH path=(a:person)-[e:knows*2..2]->(b:person) return path order by (nodes(path)[1]).ID, (nodes(path)[2]).ID, (nodes(path)[3]).ID"
        arrow_tbl = _conn.execute(query).get_as_arrow(0)
        # paths should be:
        expected_nodes = [
            [0, 2, 0],
            [0, 2, 3],
            [0, 2, 5],
            [0, 3, 0],
            [0, 3, 2],
            [0, 3, 5],
            [0, 5, 0],
            [0, 5, 2],
            [0, 5, 3],
            [2, 0, 2],
            [2, 0, 3],
            [2, 0, 5],
            [2, 3, 0],
            [2, 3, 2],
            [2, 3, 5],
            [2, 5, 0],
            [2, 5, 2],
            [2, 5, 3],
            [3, 0, 2],
            [3, 0, 3],
            [3, 0, 5],
            [3, 2, 0],
            [3, 2, 3],
            [3, 2, 5],
            [3, 5, 0],
            [3, 5, 2],
            [3, 5, 3],
            [5, 0, 2],
            [5, 0, 3],
            [5, 0, 5],
            [5, 2, 0],
            [5, 2, 3],
            [5, 2, 5],
            [5, 3, 0],
            [5, 3, 2],
            [5, 3, 5],
        ]
        for row, expected in zip(arrow_tbl["path"], expected_nodes):
            cur_ids = []
            rel_ids = []
            for node in row["_NODES"]:
                cur_ids += [node["ID"].as_py()]
            assert expected == cur_ids

    def _test_serial(_conn: kuzu.Connection) -> None:
        arrow_tbl = _conn.execute("MATCH (a:moviesSerial) RETURN a.ID AS id").get_as_arrow(0)
        assert arrow_tbl.num_columns == 1
        assert len(arrow_tbl) == 3
        assert arrow_tbl["id"].to_pylist() == [0, 1, 2]

    def _test_blob(_conn: kuzu.Connection) -> None:
        arrow_tbl = _conn.execute("RETURN BLOB('\\\\xBC\\\\xBD\\\\xBA\\\\xAA') as result").get_as_arrow(1)
        assert arrow_tbl.num_columns == 1
        assert len(arrow_tbl) == 1
        assert arrow_tbl["result"][0].as_py() == b"\xbc\xbd\xba\xaa"

    _test_node(conn)
    _test_node_rel(conn)
    _test_marries_table(conn)
    _test_recursive_rel(conn)
    _test_serial(conn)
    _test_blob(conn)

    def test_to_arrow1(conn: kuzu.Connection) -> None:
        query = "MATCH (a:person)-[e:knows]->(:person) RETURN e.summary"
        res = conn.execute(query)
        arrow_tbl = conn.execute(query).get_as_arrow(-1)  # what is a chunk size of -1 even supposed to mean?
        assert arrow_tbl == []



import pathlib
import os
from packaging.version import Version

CURRENT_DIR = pathlib.Path(__file__).parent.resolve()

RELEASES_PATH = CURRENT_DIR.joinpath('releases').resolve()

production_releases = open(CURRENT_DIR.joinpath('PRODUCTION_RELEASES')).read().splitlines()

releases_to_purge = [r for r in os.listdir(RELEASES_PATH) if r.startswith('v')]

releases_to_purge = [r for r in releases_to_purge if r not in production_releases]
releases_to_purge = [r[1:] for r in releases_to_purge]
releases_to_purge.sort(key=Version)

releases_to_purge.pop()

releases_to_purge = ['v' + r for r in releases_to_purge]

if len(releases_to_purge) == 0:
    print('No releases to purge.')
    exit(0)

print('Releases to purge:')
for r in releases_to_purge:
    print('  ' + r)

for r in releases_to_purge:
    path_to_purge = RELEASES_PATH.joinpath(r)
    print('Deleting ' + str(path_to_purge))
    os.system('rm -rf ' + str(path_to_purge))

print('Done.')



from enum import Enum


class Type(Enum):
    """The type of a value in the database."""

    ANY = "ANY"
    NODE = "NODE"
    REL = "REL"
    RECURSIVE_REL = "RECURSIVE_REL"
    SERIAL = "SERIAL"
    BOOL = "BOOL"
    INT64 = "INT64"
    INT32 = "INT32"
    INT16 = "INT16"
    INT8 = "INT8"
    UINT64 = "UINT64"
    UINT32 = "UINT32"
    UINT16 = "UINT16"
    UINT8 = "UINT8"
    INT128 = "INT128"
    DOUBLE = "DOUBLE"
    FLOAT = "FLOAT"
    DATE = "DATE"
    TIMESTAMP = "TIMESTAMP"
    TIMSTAMP_TZ = "TIMESTAMP_TZ"
    TIMESTAMP_NS = "TIMESTAMP_NS"
    TIMESTAMP_MS = "TIMESTAMP_MS"
    TIMESTAMP_SEC = "TIMESTAMP_SEC"
    INTERVAL = "INTERVAL"
    INTERNAL_ID = "INTERNAL_ID"
    STRING = "STRING"
    BLOB = "BLOB"
    UUID = "UUID"
    LIST = "LIST"
    ARRAY = "ARRAY"
    STRUCT = "STRUCT"
    MAP = "MAP"
    UNION = "UNION"



from conftest import ShellTest


def test_help(temp_db) -> None:
    test = ShellTest().add_argument(temp_db).statement(":help")
    result = test.run()
    result.check_stdout(
        [
            "    :help     get command list",
            "    :clear     clear shell",
            "    :quit     exit from shell",
            "    :max_rows [max_rows]     set maximum number of rows for display (default: 20)",
            "    :max_width [max_width]     set maximum width in characters for display",
            "    :mode [mode]     set output mode (default: box)",
            "    :stats [on|off]     toggle query stats on or off",  
            "    :multiline     set multiline mode (default)",
            "    :singleline     set singleline mode",
            "    :highlight [on|off]     toggle syntax highlighting on or off",
            "    :render_errors [on|off]     toggle error highlighting on or off",
            "    :render_completion [on|off]     toggle completion highlighting on or off",
            "",
            "    Note: you can change and see several system configurations, such as num-threads, ",
            "          timeout, and progress_bar using Cypher CALL statements.",
            "          e.g. CALL THREADS=5; or CALL current_setting('threads') return *;",
            "          See: \x1B]8;;https://docs.kuzudb.com/cypher/configuration\x1B\\https://docs.kuzudb.com/cypher/configuration\x1B]8;;\x1B\\",
        ],
    )


def test_clear(temp_db) -> None:
    test = ShellTest().add_argument(temp_db).statement(":clear")
    result = test.run()
    result.check_stdout("\x1b[H\x1b[2J")


def test_quit(temp_db) -> None:
    test = ShellTest().add_argument(temp_db).statement(":quit").statement("RETURN RANGE(0,10) AS a;")
    result = test.run()
    # check to make sure the return query did not execute
    result.check_not_stdout("[0,1,2,3,4,5,6,7,8,9,10]")


def test_max_rows(temp_db, csv_path) -> None:
    # test all rows shown
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":max_rows 40")
        .statement(f'LOAD FROM "{csv_path}" (HEADER=true) RETURN id;')
    )
    result = test.run()
    result.check_stdout("(21 tuples)")
    result.check_not_stdout(["\u2502  \u00B7    \u2502", "\u2502  \u00B7    \u2502", "\u2502  \u00B7    \u2502"])

    # test 1 row shown
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":max_rows 1")
        .statement(f'LOAD FROM "{csv_path}" (HEADER=true) RETURN id;')
    )
    result = test.run()
    result.check_stdout("(21 tuples, 1 shown)")
    result.check_stdout(["\u2502  \u00B7    \u2502", "\u2502  \u00B7    \u2502", "\u2502  \u00B7    \u2502"])

    # test setting back to default
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":max_rows 0")
        .statement(f'LOAD FROM "{csv_path}" (HEADER=true) RETURN id;')
    )
    result = test.run()
    result.check_stdout("(21 tuples, 20 shown)")
    result.check_stdout(["\u2502  \u00B7    \u2502", "\u2502  \u00B7    \u2502", "\u2502  \u00B7    \u2502"])


def test_max_width(temp_db, csv_path) -> None:
    # test all columns shown
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":max_width 400")
        .statement(f'LOAD FROM "{csv_path}" (HEADER=true) RETURN *;')
    )
    result = test.run()
    result.check_stdout("id")
    result.check_stdout("fname")
    result.check_stdout("Gender")
    result.check_stdout("ISStudent")
    result.check_stdout("isWorker")
    result.check_stdout("age")
    result.check_stdout("eyeSight")
    result.check_stdout("birthdate")
    result.check_stdout("registerTime")
    result.check_stdout("lastJobDuration")
    result.check_stdout("workedHours")
    result.check_stdout("usedNames")
    result.check_stdout("courseScoresPerTerm")
    result.check_not_stdout("\u2502 ... \u2502")
    result.check_stdout("(13 columns)")

    # test 2 columns shown
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":max_width 44")
        .statement(f'LOAD FROM "{csv_path}" (HEADER=true) RETURN *;')
    )
    result = test.run()
    result.check_stdout("id")
    result.check_not_stdout("fname")
    result.check_not_stdout("Gender")
    result.check_not_stdout("ISStudent")
    result.check_not_stdout("isWorker")
    result.check_not_stdout("\u2502 age    \u2502")
    result.check_not_stdout("eyeSight")
    result.check_not_stdout("birthdate")
    result.check_not_stdout("registerTime")
    result.check_not_stdout("lastJobDuration")
    result.check_not_stdout("workedHours")
    result.check_not_stdout("usedNames")
    result.check_stdout("\u2502 ... \u2502")
    result.check_stdout("courseScoresPerTerm")
    result.check_stdout("(13 columns, 2 shown)")

    # test too small to display (back to terminal width)
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":max_width 2")
        .statement(f'LOAD FROM "{csv_path}" (HEADER=true) RETURN *;')
    )
    result = test.run()
    # terminal width when running test is 80
    result.check_stdout("id")
    result.check_stdout("fname")
    result.check_stdout("Gender")
    result.check_not_stdout("ISStudent")
    result.check_not_stdout("isWorker")
    result.check_not_stdout("\u2502 age    \u2502")
    result.check_not_stdout("eyeSight")
    result.check_not_stdout("birthdate")
    result.check_not_stdout("registerTime")
    result.check_not_stdout("lastJobDuration")
    result.check_not_stdout("workedHours")
    result.check_not_stdout("usedNames")
    result.check_stdout("\u2502 ... \u2502")
    result.check_stdout("courseScoresPerTerm")
    result.check_stdout("(13 columns, 4 shown)")

    # test result tuples unaffected by width
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":max_width 2")
        .statement(
            "CREATE NODE TABLE LANGUAGE_CODE(alpha2 STRING, English STRING, PRIMARY KEY (alpha2));",
        )
    )
    result = test.run()
    # terminal width when running test is 80
    result.check_stdout("\u2502 Table LANGUAGE_CODE has been created. \u2502")
    result.check_not_stdout("\u2502 ... \u2502")
    result.check_stdout("(1 column)")


def output_mode_verification(result) -> None:
    result.check_stdout(
        [
            "Available output modes:",
            "    box (default):    Tables using unicode box-drawing characters",
            "    column:    Output in columns",
            "    csv:    Comma-separated values",
            "    html:    HTML table",
            "    json:    Results in a JSON array",
            "    jsonlines:    Results in a NDJSON format",
            "    latex:    LaTeX tabular environment code",
            "    line:    One value per line",
            "    list:    Values delimited by \"|\"",
            "    markdown:    Markdown table",
            "    table:    Tables using ASCII characters",
            "    tsv:    Tab-separated values",
            "    trash:    No output",
        ],
    )


def test_set_mode(temp_db) -> None:
    # test default mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("\u2502 a              \u2502 b            \u2502")
    result.check_stdout("\u2502 Databases Rule \u2502 kuzu is cool \u2502")

    # test column mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode column")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a                b")
    result.check_stdout("Databases Rule   kuzu is cool")

    # test csv mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode csv")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a,b")
    result.check_stdout("Databases Rule,kuzu is cool")

    # test csv escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode csv")
        .statement('RETURN "This is a \\"test\\", with commas, \\"quotes\\", and\nnewlines.";')
    )
    result = test.run()
    result.check_stdout('"This is a ""test"", with commas, ""quotes"", and\nnewlines."')

    # test box mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        # box is default so need to switch to another mode first
        .statement(":mode csv")
        .statement(":mode box")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("\u2502 a              \u2502 b            \u2502")
    result.check_stdout("\u2502 Databases Rule \u2502 kuzu is cool \u2502")

    # test html mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode html")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("<table>")
    result.check_stdout("<tr>")
    result.check_stdout("<th>a</th><th>b</th>")
    result.check_stdout("</tr>")
    result.check_stdout("<tr>")
    result.check_stdout("<td>Databases Rule</td><td>kuzu is cool</td>")
    result.check_stdout("</tr>")
    result.check_stdout("</table>")

    # test html escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode html")
        .statement('RETURN "This is a <test> & \\"example\\" with \'special\' characters." AS a;')
    )
    result = test.run()
    result.check_stdout("<table>")
    result.check_stdout("<tr>")
    result.check_stdout("<th>a</th>")
    result.check_stdout("</tr>")
    result.check_stdout("<tr>")
    result.check_stdout("<td>This is a &lt;test&gt; &amp; &quot;example&quot; with &apos;special&apos; characters.</td>")
    result.check_stdout("</tr>")
    result.check_stdout("</table>")

    # test json mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode json")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout('[\n{"a":"Databases Rule","b":"kuzu is cool"}\n]')

    # test json escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode json")
        .statement('RETURN "This is a \\"test\\" with backslashes \\\\, newlines\n, and tabs \t." AS a;')
    )
    result = test.run()
    result.check_stdout('[\n{"a":"This is a \\"test\\" with backslashes \\\\, newlines\\n, and tabs \\t."}\n]')

    # test jsonlines mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode jsonlines")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout('{"a":"Databases Rule","b":"kuzu is cool"}')

    # test jsonlines escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode jsonlines")
        .statement('RETURN "This is a \\"test\\" with backslashes \\\\, newlines\n, and tabs \t." AS a;')
    )
    result = test.run()
    result.check_stdout('{"a":"This is a \\"test\\" with backslashes \\\\, newlines\\n, and tabs \\t."}')

    # test latex mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode latex")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("\\begin{tabular}{ll}")
    result.check_stdout("\\hline")
    result.check_stdout("a&b\\\\")
    result.check_stdout("\\hline")
    result.check_stdout("Databases Rule&kuzu is cool\\\\")
    result.check_stdout("\\hline")
    result.check_stdout("\\end{tabular}")

    # test latex escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode latex")
        .statement('RETURN "This is a test with special characters: %, $, &, #, _, {, }, ~, ^, \\\\, <, and >." AS a;')
    )
    result = test.run()
    result.check_stdout("\\begin{tabular}{l}")
    result.check_stdout("\\hline")
    result.check_stdout("a\\\\")
    result.check_stdout("\\hline")
    result.check_stdout("This is a test with special characters: \\%, \\$, \\&, \\#, \\_, \\{, \\}, \\textasciitilde{}, \\textasciicircum{}, \\textbackslash{}, \\textless{}, and \\textgreater{}.\\\\")
    result.check_stdout("\\hline")
    result.check_stdout("\\end{tabular}")

    # test line mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode line")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a = Databases Rule")
    result.check_stdout("b = kuzu is cool")
    
    # test list mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode list")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a|b")
    result.check_stdout("Databases Rule|kuzu is cool")

    # test list escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode tsv")
        .statement('RETURN "This is a \\"test\\", with vertical bars |, \\"quotes\\", and\nnewlines.";')
    )
    result = test.run()
    result.check_stdout('"This is a ""test"", with vertical bars |, ""quotes"", and\nnewlines."')

    # test markdown mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode markdown")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("| a              | b            |")
    result.check_stdout("| Databases Rule | kuzu is cool |")

    # test table mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode table")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("| a              | b            |")
    result.check_stdout("| Databases Rule | kuzu is cool |")

    # test tsv mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode tsv")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("a\tb")
    result.check_stdout("Databases Rule\tkuzu is cool")

    # test tsv escaping
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode tsv")
        .statement('RETURN "This is a \\"test\\", with tabs \t, \\"quotes\\", and\nnewlines.";')
    )
    result = test.run()
    result.check_stdout('"This is a ""test"", with tabs \t, ""quotes"", and\nnewlines."')

    # test trash mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode trash")
        .statement('RETURN RANGE(0, 10) AS a;')
    )
    result = test.run()
    result.check_not_stdout("[0,1,2,3,4,5,6,7,8,9,10]")

    # test mode info
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode invalid")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    output_mode_verification(result)

    # test invalid mode
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":mode invalid")
        .statement('RETURN "Databases Rule" AS a, "kuzu is cool" AS b;')
    )
    result = test.run()
    result.check_stdout("Cannot parse 'invalid' as output mode.")
    output_mode_verification(result)


def test_stats(temp_db) -> None:
    # test stats default
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement('RETURN "Databases Rule" AS a;')
    )
    result = test.run()
    result.check_stdout("(1 tuple)")
    result.check_stdout("(1 column)")
    result.check_stdout("Time: ")

    # test stats off
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":stats off")
        .statement('RETURN "Databases Rule" AS a;')
    )
    result = test.run()
    result.check_not_stdout("(1 tuple)")
    result.check_not_stdout("(1 column)")
    result.check_not_stdout("Time: ")

    # test stats on
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":stats on")
        .statement('RETURN "Databases Rule" AS a;')
    )
    result = test.run()
    result.check_stdout("(1 tuple)")
    result.check_stdout("(1 column)")
    result.check_stdout("Time: ")

    # test stats invalid
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":stats invalid")
        .statement('RETURN "Databases Rule" AS a;')
    )
    result = test.run()
    result.check_stdout("Cannot parse 'invalid' to toggle stats. Expect 'on' or 'off'.")


def test_multiline(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":multiline")
        .statement("RETURN")
        .statement('"databases rule"')
        .statement("AS")
        .statement("a")
        .statement(";")
    )
    result = test.run()
    result.check_stdout("\u2502 databases rule \u2502")

    # test no truncation
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement("a" * 400 + ";")
    )
    result = test.run()
    result.check_stdout("a" * 400)


def test_singleline(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":multiline")
        .statement(":singleline")
        .statement("RETURN")
        .statement('"databases rule"')
        .statement("AS")
        .statement("a")
        .statement(";")
    )
    result = test.run()
    result.check_stdout("\u2502 databases rule \u2502")

    # test truncation
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement("a" * 400 + ";")
    )
    result = test.run()
    result.check_stdout("a" * 80)


def test_highlight(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":highlight off")
    )
    result = test.run()
    result.check_stdout("disabled syntax highlighting")

    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":highlight on")
    )
    result = test.run()
    result.check_stdout("enabled syntax highlighting")

    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":highlight o")
    )
    result = test.run()
    result.check_stdout("Cannot parse 'o' to toggle highlighting. Expect 'on' or 'off'.")


def test_render_errors(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":render_errors off")
    )
    result = test.run()
    result.check_stdout("disabled error highlighting")

    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":render_errors on")
    )
    result = test.run()
    result.check_stdout("enabled error highlighting")

    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":render_errors o")
    )
    result = test.run()
    result.check_stdout("Cannot parse 'o' to toggle error highlighting. Expect 'on' or 'off'.")


def test_completion_highlighting(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":render_completion off")
    )
    result = test.run()
    result.check_stdout("disabled completion highlighting")

    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":render_completion on")
    )
    result = test.run()
    result.check_stdout("enabled completion highlighting")

    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement(":render_completion o")
    )
    result = test.run()
    result.check_stdout("Cannot parse 'o' to toggle completion highlighting. Expect 'on' or 'off'.")


def test_bad_command(temp_db) -> None:
    test = ShellTest().add_argument(temp_db).statement(":maxrows").statement(":quiy").statement("clearr;")
    result = test.run()
    result.check_stdout(
        'Error: Unknown command: ":maxrows". Enter ":help" for help',
    )
    result.check_stdout('Did you mean: ":max_rows"?')
    result.check_stdout('Error: Unknown command: ":quiy". Enter ":help" for help')
    result.check_stdout('Did you mean: ":quit"?')
    result.check_stdout(
        '"clearr;" is not a valid Cypher query. Did you mean to issue a CLI command, e.g., ":clear"?',
    )



from __future__ import annotations

import kuzu

from type_aliases import ConnDB


def test_iteration_list(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2

    rows = list(result)
    assert rows[0] == ["Alice", 30]
    assert rows[1] == ["Bob", 40]


def test_iteration_loop(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2

    rows = []
    for row in result:
        rows.append(row)

    assert rows[0] == ["Alice", 30]
    assert rows[1] == ["Bob", 40]


def test_get_all(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    db = kuzu.Database(database_path=":memory:")
    assert not db.is_closed
    assert db._database is not None

    conn = kuzu.Connection(db)
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2

    rows = result.get_all()

    assert rows[0] == ["Alice", 30]
    assert rows[1] == ["Bob", 40]


def test_get_n(conn_db_in_mem: ConnDB) -> None:
    conn, _ = conn_db_in_mem
    db = kuzu.Database(database_path=":memory:")
    assert not db.is_closed
    assert db._database is not None

    conn = kuzu.Connection(db)
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    conn.execute("CREATE (:person {name: 'Cole', age: 20});")

    # Basic
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 3

    rows = result.get_n(2)

    assert len(rows) == 2
    assert rows[0] == ["Alice", 30]
    assert rows[1] == ["Bob", 40]

    rows = result.get_n(1)

    assert len(rows) == 1
    assert rows[0] == ["Cole", 20]

    rows = result.get_n(3)

    assert len(rows) == 0

    # Empty rows on excess count
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 3

    rows = result.get_n(10)

    assert len(rows) == 3
    assert rows[0] == ["Alice", 30]
    assert rows[1] == ["Bob", 40]
    assert rows[2] == ["Cole", 20]



import sys

from kuzu import Connection, Database

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

ConnDB: TypeAlias = tuple[Connection, Database]



import argparse
import os
import sys
import subprocess


# Parse schema.cypher and copy.cypher files.
def create_cypher_queries(file_path):
    commands = []
    try:
        with open(file_path, "r") as f:
            for line in f:
                stripped = line.strip()
                if not stripped:
                    continue
                if not stripped.endswith(";"):
                    stripped += ";"
                commands.append(stripped)
    except Exception:
        pass
    return commands


# Find all datasets that have a schema.cypher and copy.cypher file.
def find_valid_dataset_dirs(dataset_root):
    valid_dirs = []

    for root, dirs, files in os.walk(dataset_root):
        # This script creates a tmp directory with the exported dbs, we should
        # skip it in our search.
        if "tmp" in root.split(os.sep):
            continue
        file_set = set(files)

        if "schema.cypher" in file_set:
            valid_dirs.append(root)

    return valid_dirs


# Example scripts/export-dbs.py build/debug/tools/shell/kuzu dataset.
def main():
    parser = argparse.ArgumentParser(
        description="Export DBs with KUZU shell and dataset paths"
    )

    parser.add_argument(
        "--executable", required=True, help="Path to the KUZU shell executable"
    )
    parser.add_argument(
        "--dataset-dir", required=True, help="Path to the dataset directory"
    )
    parser.add_argument(
        "--output-dir", required=True, help="Path to export the datasets"
    )
    args = parser.parse_args()

    arg_executable_path = os.path.abspath(args.executable)
    arg_dataset_path = os.path.abspath(args.dataset_dir)
    output_dir = os.path.abspath(args.output_dir)

    if not os.path.isfile(arg_executable_path):
        raise Exception(f"Error: Executable not found at {arg_executable_path}")
    if not os.path.exists(arg_dataset_path):
        raise Exception(f"Error: Dataset path not found at {arg_dataset_path}")

    valid_datasets = find_valid_dataset_dirs(arg_dataset_path)
    # This is done to construct a full path to replace the relative paths found
    # in copy.cypher files.
    script_dir = os.path.dirname(os.path.realpath(__file__))
    root_dir = os.path.abspath(os.path.join(script_dir, ".."))
    for dataset_path in valid_datasets:
        schema_commands = create_cypher_queries(
            os.path.join(dataset_path, "schema.cypher")
        )
        copy_commands = create_cypher_queries(os.path.join(dataset_path, "copy.cypher"))
        combined_commands = schema_commands + copy_commands
        dataset_name = os.path.relpath(dataset_path, arg_dataset_path)
        export_path = os.path.join(output_dir, dataset_name)
        export_command = (
            f"EXPORT DATABASE '{export_path}' (format=\"csv\", header=true);"
        )
        combined_commands.append(export_command)
        combined_commands.insert(0, "CALL threads=1;")
        print(f"Exporting {dataset_path} to {export_path}")
        joined_commands = "\n".join(cmd.strip() for cmd in combined_commands)

        subprocess.run(
            arg_executable_path,
            input=joined_commands,
            text=True,
            cwd=root_dir,
            check=True,
            shell=True,
        )

    return 0


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

TINY_SNB_KNOWS_GROUND_TRUTH = {
    0: [2, 3, 5],
    2: [0, 3, 5],
    3: [0, 2, 5],
    5: [0, 2, 3],
    7: [8, 9],
}

TINY_SNB_PERSON_IDS_GROUND_TRUTH = [0, 2, 3, 5, 7, 8, 9, 10]

# TODO: FIX-ME. Re-enable the tests when StorageDriver is fixed.
# def test_remote_backend_graph_store(conn_db_readonly: ConnDB) -> None:
#     _, db = conn_db_readonly
#     fs, gs = db.get_torch_geometric_remote_backend()
#     edge_idx = gs.get_edge_index(("person", "knows", "person"), layout="coo", is_sorted=False)
#     assert edge_idx.shape == torch.Size([2, 14])
#     for i in range(14):
#         src = edge_idx[0, i].item()
#         src_id = fs["person", "ID", src].item()
#         dst = edge_idx[1, i].item()
#         dst_id = fs["person", "ID", dst].item()
#         assert src_id in TINY_SNB_KNOWS_GROUND_TRUTH
#         assert dst_id in TINY_SNB_KNOWS_GROUND_TRUTH[src_id]
#
#
# def test_remote_backend_feature_store_idx(conn_db_readonly: ConnDB) -> None:
#     _, db = conn_db_readonly
#     fs, _ = db.get_torch_geometric_remote_backend()
#     id_set = set(TINY_SNB_PERSON_IDS_GROUND_TRUTH)
#     index_to_id = {}
#     # Individual index access
#     for i in range(len(id_set)):
#         assert fs["person", "ID", i].item() in id_set
#         id_set.remove(fs["person", "ID", i].item())
#         index_to_id[i] = fs["person", "ID", i].item()
#
#     # Range index access
#     for i in range(len(index_to_id) - 3):
#         ids = fs["person", "ID", i : i + 3]
#         for j in range(3):
#             assert ids[j].item() == index_to_id[i + j]
#
#     # Multiple index access
#     indicies = random.sample(range(len(index_to_id)), 4)
#     ids = fs["person", "ID", indicies]
#     for i in range(4):
#         idx = indicies[i]
#         assert ids[i].item() == index_to_id[idx]
#
#     # No index access
#     ids = fs["person", "ID", None]
#     assert len(ids) == len(index_to_id)
#     for i in range(len(ids)):
#         assert ids[i].item() == index_to_id[i]
#
#
# def test_remote_backend_feature_store_types_1d(conn_db_readonly: ConnDB) -> None:
#     _, db = conn_db_readonly
#     fs, _ = db.get_torch_geometric_remote_backend()
#     for i in range(3):
#         assert fs["npyoned", "i64", i].item() - i == 1
#         assert fs["npyoned", "i32", i].item() - i == 1
#         assert fs["npyoned", "i16", i].item() - i == 1
#         assert fs["npyoned", "f64", i].item() - i - 1 < 1e-6
#         assert fs["npyoned", "f32", i].item() - i - 1 < 1e-6
#
#
# def test_remote_backend_feature_store_types_2d(conn_db_readonly: ConnDB) -> None:
#     _, db = conn_db_readonly
#     fs, _ = db.get_torch_geometric_remote_backend()
#     for i in range(3):
#         i64 = fs["npytwod", "i64", i]
#         assert i64.shape == torch.Size([1, 3])
#         base_number = (i * 3) + 1
#         for j in range(3):
#             assert i64[0, j].item() == base_number + j
#         i32 = fs["npytwod", "i32", i]
#         assert i32.shape == torch.Size([1, 3])
#         for j in range(3):
#             assert i32[0, j].item() == base_number + j
#         i16 = fs["npytwod", "i16", i]
#         assert i16.shape == torch.Size([1, 3])
#         for j in range(3):
#             assert i16[0, j].item() == base_number + j
#         f64 = fs["npytwod", "f64", i]
#         assert f64.shape == torch.Size([1, 3])
#         for j in range(3):
#             assert f64[0, j].item() - (base_number + j) - 1 < 1e-6
#         f32 = fs["npytwod", "f32", i]
#         assert f32.shape == torch.Size([1, 3])
#         for j in range(3):
#             assert f32[0, j].item() - (base_number + j) - 1 < 1e-6
#
#
# def test_remote_backend_20k(conn_db_readwrite: ConnDB) -> None:
#     _, db = conn_db_readwrite
#     conn = kuzu.Connection(db, num_threads=1)
#     conn.execute("CREATE NODE TABLE npy20k (id INT64,f32 FLOAT[10],PRIMARY KEY(id));")
#     conn.execute(
#         f"""
#         COPY npy20k FROM (
#           "{KUZU_ROOT}/dataset/npy-20k/id_int64.npy",
#           "{KUZU_ROOT}/dataset/npy-20k/two_dim_float.npy") BY COLUMN;
#         """
#     )
#     del conn
#     fs, _ = db.get_torch_geometric_remote_backend(8)
#     for i in range(20000):
#         assert fs["npy20k", "id", i].item() == i



from __future__ import annotations

from datetime import date, datetime, timedelta  # noqa: TC003

import pandas as pd
import pyarrow as pa
import pytest
from kuzu import Type
from type_aliases import ConnDB


def udf_helper(conn, functionName, params, expectedResult):
    plist = ", ".join(f"${i + 1}" for i in range(len(params)))
    queryString = f"RETURN {functionName}({plist})"
    result = conn.execute(queryString, {str(i + 1): params[i] for i in range(len(params))})
    assert result.get_next()[0] == expectedResult


def udf_predicate_test(conn, functionName, params, expectedResult):
    conn.execute("CREATE node table tbl (id INT64, primary key (id))")
    for i in params:
        conn.execute("CREATE (t:tbl {id: $num})", {"num": i})
    result = conn.execute(f"MATCH (t:tbl) WHERE {functionName}(t.id % 8) RETURN COUNT(*)")
    assert result.get_next()[0] == expectedResult


def test_udf(conn_db_readwrite: ConnDB) -> None:
    conn, _ = conn_db_readwrite

    def add5int(x: int) -> int:
        print(x)
        return x + 5

    add5IntArgs = ["add5int", add5int]

    def add5float(x: float) -> float:
        return x + 5.0

    add5FloatArgs = ["add5float", add5float]

    def intToString(x: int) -> str:
        return str(x)

    intToStringArgs = ["intToString", intToString]

    def sumOf7(a, b, c, d, e, f, g) -> int:
        return a + b + c + d + e + f + g

    sumOf7Args = ["sumOf7", sumOf7, [Type.INT64] * 7]

    def dateToDatetime(a: date) -> datetime:
        return datetime(a.year, a.month, a.day)

    dateToDatetimeArgs = ["dateToDatetime", dateToDatetime]

    def datetimeToDate(a: datetime) -> date:
        return a.date()

    datetimeToDateArgs = ["datetimeToDate", datetimeToDate, [Type.TIMESTAMP], Type.DATE]

    def addToDate(a: datetime, b: timedelta) -> datetime:
        return a + b

    addToDateArgs = ["addToDate", addToDate, [Type.TIMESTAMP, Type.INTERVAL], Type.TIMESTAMP]

    def concatThreeLists(a: list[int], b: list[int], c: list[int]) -> list[int]:
        return a + b + c

    concatThreeListsArgs = ["concatThreeLists", concatThreeLists]

    def mergeMaps(a, b):
        return a | b

    mergeMapsArgs = ["mergeMaps", mergeMaps, ["MAP(STRING, INT64)"] * 2, "MAP(STRING, INT64)"]

    def mergeMaps2(a: dict[str, int], b: dict[str, int]) -> dict[str, int]:
        return a | b

    mergeMaps2Args = ["mergeMaps2", mergeMaps2]

    def selectIfSeven(a: int) -> bool:
        return a == 7

    selectIfSevenArgs = ["selectIfSeven", selectIfSeven]

    def distancePrimer(pointA, pointB):
        return {"x": (pointA["x"] - pointB["x"]) ** 2, "y": (pointA["y"] - pointB["y"]) ** 2}

    distanceArgs = [
        "UDFDist",
        distancePrimer,
        ["STRUCT(x INT32, y INT32)", "STRUCT(x INT32, y INT32)"],
        "STRUCT(x INT32, y INT32)",
    ]

    conn.create_function(*add5IntArgs)
    conn.create_function(*add5FloatArgs)
    conn.create_function(*intToStringArgs)
    conn.create_function(*sumOf7Args)
    conn.create_function(*dateToDatetimeArgs)
    conn.create_function(*datetimeToDateArgs)
    conn.create_function(*addToDateArgs)
    conn.create_function(*concatThreeListsArgs)
    conn.create_function(*mergeMapsArgs)
    conn.create_function(*mergeMaps2Args)
    conn.create_function(*selectIfSevenArgs)
    conn.create_function(*distanceArgs)

    udf_helper(conn, "add5int", [10], 15)
    udf_helper(conn, "add5int", [9], 14)
    udf_helper(conn, "add5int", [1283], 1288)
    udf_helper(conn, "add5float", [2.2], 7.2)
    udf_helper(conn, "intToString", [123], "123")
    udf_helper(conn, "sumOf7", [1, 2, 3, 4, 5, 6, 7], sum(range(8)))
    udf_helper(conn, "dateToDatetime", [date(2024, 4, 25)], datetime(2024, 4, 25))
    udf_helper(conn, "datetimeToDate", [datetime(2024, 4, 25, 15, 39, 20)], date(2024, 4, 25))
    udf_helper(
        conn,
        "addToDate",
        [datetime(2024, 4, 25, 15, 39, 20), datetime(2025, 5, 26) - datetime(2024, 4, 25, 15, 39, 20)],
        datetime(2025, 5, 26),
    )
    udf_helper(conn, "concatThreeLists", [[1, 2, 3], [4, 5, 6], [-1, -2, -3]], [1, 2, 3, 4, 5, 6, -1, -2, -3])
    udf_helper(
        conn,
        "mergeMaps",
        [{"key": ["a", "b", "c"], "value": [1, 2, 3]}, {"key": ["x", "y", "z"], "value": [100, 200, 300]}],
        {"a": 1, "b": 2, "c": 3, "x": 100, "y": 200, "z": 300},
    )
    udf_helper(
        conn,
        "mergeMaps2",
        [{"key": ["a", "b", "c"], "value": [1, 2, 3]}, {"key": ["x", "y", "z"], "value": [100, 200, 300]}],
        {"a": 1, "b": 2, "c": 3, "x": 100, "y": 200, "z": 300},
    )
    udf_predicate_test(conn, "selectIfSeven", range(65), 8)
    udf_helper(conn, "UDFDist", [{"x": 1, "y": 10}, {"x": 4, "y": 6}], {"x": 9, "y": 16})

    df = pd.DataFrame({"col": list(range(5000))})
    result = conn.execute("LOAD FROM df RETURN add5int(col) as ans").get_as_df()
    assert list(result["ans"]) == list(map(add5int, range(5000)))

    df = pd.DataFrame({
        "col1": pd.Series(
            [{"a": 1, "b": 2, "c": 3}, {"l": 1, "m": 2, "n": 3}], dtype=pd.ArrowDtype(pa.map_(pa.string(), pa.int64()))
        ),
        "col2": pd.Series(
            [{"x": -1, "y": -2, "z": -3}, {"one": -1, "two": -2, "three": -3}],
            dtype=pd.ArrowDtype(pa.map_(pa.string(), pa.int64())),
        ),
    })
    # result = conn.execute("LOAD FROM df RETURN mergeMaps(col1, col2) as ans").get_as_arrow()
    # assert result["ans"].to_pylist() == [
    #     [("a", 1), ("b", 2), ("c", 3), ("x", -1), ("y", -2), ("z", -3)],
    #     [("l", 1), ("m", 2), ("n", 3), ("one", -1), ("two", -2), ("three", -3)],
    # ]


def test_udf_null(conn_db_readwrite: ConnDB) -> None:
    conn, _ = conn_db_readwrite

    def get5(x: int) -> int:
        return 5

    conn.create_function("get5", get5)
    assert conn.execute("RETURN get5(NULL)").get_next() == [None]

    conn.remove_function("get5")
    conn.create_function("get5", get5, default_null_handling=False)
    assert conn.execute("RETURN get5(NULL)").get_next() == [5]


def test_udf_except(conn_db_readwrite: ConnDB) -> None:
    class TestException(Exception):
        pass

    conn, _ = conn_db_readwrite

    def throw() -> int:
        errmsg = "test"
        raise TestException(errmsg)

    conn.create_function("testexcept", throw)

    pytest.raises(RuntimeError, conn.execute, "RETURN testexcept()")

    conn.remove_function("testexcept")
    conn.create_function("testexcept", throw, catch_exceptions=True)

    assert conn.execute("RETURN testexcept()").get_next() == [None]


def test_udf_remove(conn_db_readwrite: ConnDB) -> None:
    conn, _ = conn_db_readwrite

    def myfunction() -> int:
        return 1

    conn.create_function("myfunction", myfunction)

    with pytest.raises(RuntimeError, match=r"Catalog exception: function notmyfunction doesn't exist."):
        conn.remove_function("notmyfunction")

    conn.remove_function("myfunction")

    with pytest.raises(RuntimeError, match=r"Catalog exception: function list_create doesn't exist."):
        conn.remove_function("list_create")



import os
import shutil
import sys
import platform
import requests

def extract_extension_version() -> str:
    import re
    pattern = r'-DKUZU_EXTENSION_VERSION=(["\'])(.*?)\1'
    with open('./CMakeLists.txt', 'r') as f:
        for line in f:
            match = re.search(pattern, line)
            if match:
                version = match.group(2)
                return version
    raise Exception("Failed to infer KUZU extension version from CMAKE file.")


def get_os() -> str:
    os_name = "linux"
    if sys.platform.startswith("win32"):
        os_name = "win"
    elif sys.platform.startswith("darwin"):
        os_name = "osx"
    return os_name


def get_arch() -> str:
    machine = platform.machine().lower()
    if "aarch64" in machine or "arm64" in machine:
        return "arm64"
    return "amd64"  # Default to amd64


extension_repo_path = 'extension/repo'
kuzu_version = 'v' + extract_extension_version()
arch_version = f"{get_os()}_{get_arch()}"
shutil.rmtree(extension_repo_path, ignore_errors=True)
os.mkdir(extension_repo_path)
path = os.path.join(extension_repo_path, kuzu_version)
os.mkdir(path)
path = os.path.join(path, arch_version)
os.mkdir(path)

# Create common dir
extensions = [d for d in os.listdir('extension') if os.path.isdir(os.path.join('extension', d)) and d != 'repo']
for extension in extensions:
    shutil.copytree(os.path.join('extension', extension, 'build'), os.path.join(path, extension))

# Download duckdb lib from official repo to local
if get_os() != "win":
    os.mkdir(os.path.join(path, "common"))
    file_name = 'libduckdb'
    if get_os() == "osx":
        file_name += '.dylib'
    else:
        file_name += '.so'
    official_repo = f"https://extension.kuzudb.com/{kuzu_version}/{arch_version}/common/{file_name}"
    response = requests.get(official_repo)
    if response.status_code == 200:
        with open(os.path.join(path, 'common', file_name), "wb") as f:
            f.write(response.content)
    else:
        print(official_repo)
        raise Exception("Failed to download shared duckdb library from official repo.")

os.system("python3 -m RangeHTTPServer 80")



from __future__ import annotations

from type_aliases import ConnDB


def test_get_column_names(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    with conn.execute("MATCH (a:person)-[e:knows]->(b:person) RETURN a.fName, e.date, b.ID;") as result:
        column_names = result.get_column_names()
        assert column_names[0] == "a.fName"
        assert column_names[1] == "e.date"
        assert column_names[2] == "b.ID"


def test_get_column_data_types(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    with conn.execute(
        "MATCH (p:person) RETURN p.ID, p.fName, p.isStudent, p.eyeSight, p.birthdate, p.registerTime, "
        "p.lastJobDuration, p.workedHours, p.courseScoresPerTerm;"
    ) as result:
        column_data_types = result.get_column_data_types()
        assert column_data_types[0] == "INT64"
        assert column_data_types[1] == "STRING"
        assert column_data_types[2] == "BOOL"
        assert column_data_types[3] == "DOUBLE"
        assert column_data_types[4] == "DATE"
        assert column_data_types[5] == "TIMESTAMP"
        assert column_data_types[6] == "INTERVAL"
        assert column_data_types[7] == "INT64[]"
        assert column_data_types[8] == "INT64[][]"


def test_get_schema(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    with conn.execute(
        "MATCH (p:person) RETURN p.ID, p.fName, p.isStudent, p.eyeSight, p.birthdate, p.registerTime, "
        "p.lastJobDuration, p.workedHours, p.courseScoresPerTerm;"
    ) as result:
        assert result.get_schema() == {
            "p.ID": "INT64",
            "p.fName": "STRING",
            "p.isStudent": "BOOL",
            "p.eyeSight": "DOUBLE",
            "p.birthdate": "DATE",
            "p.registerTime": "TIMESTAMP",
            "p.lastJobDuration": "INTERVAL",
            "p.workedHours": "INT64[]",
            "p.courseScoresPerTerm": "INT64[][]",
        }



from datetime import datetime

import polars as pl
import pytest
from type_aliases import ConnDB


def test_polars_basic(conn_db_empty: ConnDB) -> None:
    conn, _ = conn_db_empty
    df = pl.DataFrame([
        pl.Series("col1", [1, 2], dtype=pl.Int64),
        pl.Series("col2", ["a", "b"], dtype=pl.String),
        pl.Series("col3", [1.0, 2.0], dtype=pl.Float64),
        pl.Series("col4", [datetime(2020, 5, 5), datetime(2000, 1, 1)], dtype=pl.Datetime),
    ])
    result = conn.execute("LOAD FROM df RETURN *").get_as_pl()
    equivalency = result == df
    assert equivalency["col1"].all()
    assert equivalency["col2"].all()
    assert equivalency["col3"].all()
    assert equivalency["col4"].all()
    assert result["col1"].dtype == pl.Int64
    assert result["col2"].dtype == pl.String
    assert result["col3"].dtype == pl.Float64
    assert result["col4"].dtype == pl.Datetime
    # since polars just uses arrow as its backend, it's probably not necessary to extensively test this functionality
    conn.execute(
        "CREATE NODE TABLE polarstab(col1 INT64, col2 STRING, col3 DOUBLE, col4 TIMESTAMP, PRIMARY KEY(col1))"
    )
    conn.execute("COPY polarstab FROM df")
    result = conn.execute(
        "MATCH (t:polarstab) RETURN t.col1 AS col1, t.col2 AS col2, t.col3 AS col3, t.col4 AS col4 ORDER BY col1"
    ).get_as_pl()
    equivalency = result == df
    assert equivalency["col1"].all()
    assert equivalency["col2"].all()
    assert equivalency["col3"].all()
    assert equivalency["col4"].all()
    assert result["col1"].dtype == pl.Int64
    assert result["col2"].dtype == pl.String
    assert result["col3"].dtype == pl.Float64
    assert result["col4"].dtype == pl.Datetime


def test_polars_basic_param(conn_db_empty: ConnDB) -> None:
    conn, _ = conn_db_empty
    df = pl.DataFrame([
        pl.Series("col1", [1, 2], dtype=pl.Int64),
        pl.Series("col2", ["a", "b"], dtype=pl.String),
        pl.Series("col3", [1.0, 2.0], dtype=pl.Float64),
        pl.Series("col4", [datetime(2020, 5, 5), datetime(2000, 1, 1)], dtype=pl.Datetime),
    ])
    result = conn.execute("LOAD FROM df RETURN *").get_as_pl()
    equivalency = result == df
    assert equivalency["col1"].all()
    assert equivalency["col2"].all()
    assert equivalency["col3"].all()
    assert equivalency["col4"].all()
    assert result["col1"].dtype == pl.Int64
    assert result["col2"].dtype == pl.String
    assert result["col3"].dtype == pl.Float64
    assert result["col4"].dtype == pl.Datetime
    # since polars just uses arrow as its backend, it's probably not necessary to extensively test this functionality
    conn.execute(
        "CREATE NODE TABLE polarstab(col1 INT64, col2 STRING, col3 DOUBLE, col4 TIMESTAMP, PRIMARY KEY(col1))"
    )
    conn.execute("COPY polarstab FROM $df", {"df": df})
    result = conn.execute(
        "MATCH (t:polarstab) RETURN t.col1 AS col1, t.col2 AS col2, t.col3 AS col3, t.col4 AS col4 ORDER BY col1"
    ).get_as_pl()
    equivalency = result == df
    assert equivalency["col1"].all()
    assert equivalency["col2"].all()
    assert equivalency["col3"].all()
    assert equivalency["col4"].all()
    assert result["col1"].dtype == pl.Int64
    assert result["col2"].dtype == pl.String
    assert result["col3"].dtype == pl.Float64
    assert result["col4"].dtype == pl.Datetime


def test_polars_error(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    with pytest.raises(RuntimeError, match=r"Binder exception: Variable df is not in scope."):
        conn.execute("LOAD FROM df RETURN *;")
    df = []
    with pytest.raises(
        RuntimeError,
        match=r"Binder exception: Attempted to scan from unsupported python object. Can only scan from pandas/polars dataframes and pyarrow tables.",
    ):
        conn.execute("LOAD FROM df RETURN *;")


def test_polars_scan_ignore_errors(conn_db_empty: ConnDB) -> None:
    conn, _ = conn_db_empty
    df = pl.DataFrame({"id": [1, 2, 3, 1]})
    conn.execute("CREATE NODE TABLE ids(id INT64, PRIMARY KEY(id))")
    conn.execute("COPY ids FROM df(IGNORE_ERRORS=true)")

    people = conn.execute("MATCH (i:ids) RETURN i.id")
    assert people.get_next() == [1]
    assert people.get_next() == [2]
    assert people.get_next() == [3]
    assert not people.has_next()

    warnings = conn.execute("CALL show_warnings() RETURN *")
    assert warnings.get_next()[1].startswith("Found duplicated primary key value 1")
    assert not warnings.has_next()


def test_copy_from_polars_multi_pairs(conn_db_empty: ConnDB) -> None:
    conn, _ = conn_db_empty
    conn.execute("CREATE NODE TABLE prof(id INT64, PRIMARY KEY(id))")
    conn.execute("CREATE (p:prof {id: 3});")
    conn.execute("CREATE (p:prof {id: 4});")
    conn.execute("CREATE NODE TABLE student(id INT64, PRIMARY KEY(id))")
    conn.execute("CREATE (p:student {id: 2});")
    conn.execute("CREATE REL TABLE teaches(from prof to prof, from prof to student, length int64)")
    df = pl.DataFrame({"from": [3], "to": [4], "length": [252]})
    conn.execute("COPY teaches from df (from = 'prof', to = 'prof');")
    result = conn.execute("match (:prof)-[e:teaches]->(:prof) return e.*")
    assert result.has_next()
    assert result.get_next()[0] == 252
    assert not result.has_next()


def test_scan_from_empty_lst(conn_db_empty: ConnDB) -> None:
    conn, _ = conn_db_empty
    df = pl.DataFrame({"prop1": [3], "prop2": [[]]})
    result = conn.execute("LOAD FROM df RETURN *")
    assert result.has_next()
    tp = result.get_next()
    assert tp[0] == 3
    assert tp[1] == []


def test_scan_from_parameterized_df_docs_example_1(conn_db_empty: ConnDB):
    conn, _ = conn_db_empty

    conn.execute("CREATE NODE TABLE Person(name STRING, age INT64, PRIMARY KEY (name))")

    df = pl.DataFrame({"name": ["Adam", "Karissa", "Zhang"], "age": [30, 40, 50]})

    conn.execute("COPY Person FROM $dataframe", {"dataframe": df})


def test_scan_from_parameterized_df_docs_example_2(conn_db_empty: ConnDB):
    conn, _ = conn_db_empty

    conn.execute("CREATE NODE TABLE Person(name STRING, age INT64, PRIMARY KEY (name))")

    def get_df():
        return pl.DataFrame({"name": ["Adam", "Karissa", "Zhang"], "age": [30, 40, 50]})

    conn.execute("COPY Person FROM $dataframe", {"dataframe": get_df()})


def test_scan_from_df_docs_example(conn_db_empty: ConnDB):
    conn, _ = conn_db_empty

    conn.execute("CREATE NODE TABLE Person(name STRING, age INT64, PRIMARY KEY (name))")

    df = pl.DataFrame({"name": ["Adam", "Karissa", "Zhang"], "age": [30, 40, 50]})

    conn.execute("COPY Person FROM df")



# Step 04 - generate Java literals.
#
# Java byte-code has severe restrictions. There is no such thing as
# "array literal" - those are implemented as series of data[x] = y;
# as a consequence N-byte array will use 7N bytes in class, plus N bytes
# in instantiated variable. Also no literal could be longer than 64KiB.
#
# To keep dictionary data compact both in source code and in compiled format
# we use the following tricks:
#  * use String as a data container
#  * store only lowest 7 bits; i.e. all characters fit ASCII table; this allows
#    efficient conversion to byte array; also ASCII characters use only 1 byte
#.   of memory (UTF-8 encoding)
#  * RLE-compress sequence of 8-th bits
#
# This script generates literals used in Java code.

try:
  unichr  # Python 2
except NameError:
  unichr = chr  # Python 3

bin_path = "dictionary.bin"

with open(bin_path, "rb") as raw:
  data = raw.read()

low = []
hi = []
is_skip = True
skip_flip_offset = 36
cntr = skip_flip_offset
for b in data:
  value = ord(b)
  low.append(chr(value & 0x7F))
  if is_skip:
    if value < 0x80:
      cntr += 1
    else:
      is_skip = False
      hi.append(unichr(cntr))
      cntr = skip_flip_offset + 1
  else:
    if value >= 0x80:
      cntr += 1
    else:
      is_skip = True
      hi.append(unichr(cntr))
      cntr = skip_flip_offset + 1
hi.append(unichr(cntr))

low0 = low[0:len(low) // 2]
low1 = low[len(low) // 2:len(low)]


def escape(chars):
  result = []
  for c in chars:
    if "\r" == c:
      result.append("\\r")
    elif "\n" == c:
      result.append("\\n")
    elif "\t" == c:
      result.append("\\t")
    elif "\"" == c:
      result.append("\\\"")
    elif "\\" == c:
      result.append("\\\\")
    elif ord(c) < 32 or ord(c) >= 127:
      result.append("\\u%04X" % ord(c))
    else:
      result.append(c)
  return result


source_code = [
    "  private static final String DATA0 = \"", "".join(escape(low0)), "\";\n",
    "  private static final String DATA1 = \"", "".join(escape(low1)), "\";\n",
    "  private static final String SKIP_FLIP = \"", "".join(escape(hi)), "\";\n"
]

src_path = "DictionaryData.inc.java"

with open(src_path, "w") as source:
  source.write("".join(source_code))



from pathlib import Path

import kuzu
import pytest
from test_helper import KUZU_ROOT
from conftest import get_db_file_path


def get_used_page_ranges(conn, table, column=None):
    used_pages = []
    if column is None:
        storage_info = conn.execute(f'call storage_info("{table}") return start_page_idx, num_pages')
    else:
        storage_info = conn.execute(
            f'call storage_info("{table}") where prefix(column_name, "{column}") return start_page_idx, num_pages'
        )
    while storage_info.has_next():
        cur_tuple = storage_info.get_next()
        # don't add empty (e.g. constant-compressed) pages
        if cur_tuple[1] > 0:
            used_pages.append(cur_tuple)
    return used_pages


def get_total_used_pages(conn):
    return conn.execute("call file_info() return num_pages").get_next()[0]


def get_free_page_ranges(conn):
    free_pages = []
    result = conn.execute("call fsm_info() return *")
    while result.has_next():
        cur_tuple = result.get_next()
        assert cur_tuple[1] > 0
        free_pages.append(cur_tuple)
    return free_pages


def combine_adjacent_page_ranges(page_ranges):
    new_page_ranges = []
    last_range = None
    for page_range in page_ranges:
        if last_range is not None and last_range[0] + last_range[1] == page_range[0]:
            new_range = [last_range[0], last_range[1] + page_range[1]]
            last_range = new_range
        else:
            if last_range is not None:
                new_page_ranges.append(last_range)
            last_range = page_range
    if last_range:
        new_page_ranges.append(last_range)
    return new_page_ranges


def compare_page_range_lists(used_list, free_list):
    used_list.sort()
    free_list.sort()
    # used pages should be subset of new free pages
    for entry in combine_adjacent_page_ranges(used_list):
        assert any(entry[0] >= i[0] and entry[0] + entry[1] <= i[0] + i[1] for i in free_list)


def prevent_data_file_truncation(conn):
    conn.execute("CREATE NODE TABLE IF NOT EXISTS TMP(ID INT64, PRIMARY KEY(ID))")
    conn.execute("CREATE (:TMP{id: 1})")
    conn.execute("CREATE (:TMP{id: 2})")


@pytest.fixture
def fsm_node_table_setup(tmp_path: Path):
    db = kuzu.Database(get_db_file_path(tmp_path))
    conn = kuzu.Connection(db)
    conn.execute("call threads=1")
    conn.execute("call auto_checkpoint=false")
    conn.execute(
        "create node table person (ID INt64, fName StRING, gender INT64, isStudent BoOLEAN, isWorker BOOLEAN, age INT64, eyeSight DOUBLE, birthdate DATE, registerTime TIMESTAMP, lastJobDuration interval, workedHours INT64[], usedNames STRING[], courseScoresPerTerm INT64[][], grades INT64[4], height float, u UUID, PRIMARY KEY (ID));"
    )
    conn.execute(
        f"COPY person FROM ['{KUZU_ROOT}/dataset/tinysnb/vPerson.csv', '{KUZU_ROOT}/dataset/tinysnb/vPerson2.csv'](ignore_errors=true, header=false)"
    )
    return db, conn


@pytest.fixture
def fsm_rel_table_setup(fsm_node_table_setup):
    _, conn = fsm_node_table_setup
    conn.execute(
        "create rel table knows (FROM person TO person, date DATE, meetTime TIMESTAMP, validInterval INTERVAL, comments STRING[], summary STRUCT(locations STRING[], transfer STRUCT(day DATE, amount INT64[])), notes UNION(firstmet DATE, type INT16, comment STRING), someMap MAP(STRING, STRING), MANY_MAnY);"
    )
    conn.execute(
        f"COPY knows FROM ['{KUZU_ROOT}/dataset/tinysnb/eKnows.csv', '{KUZU_ROOT}/dataset/tinysnb/eKnows_2.csv']"
    )
    return fsm_node_table_setup


@pytest.fixture
def fsm_rel_group_setup(tmp_path: Path):
    db = kuzu.Database(get_db_file_path(tmp_path))
    conn = kuzu.Connection(db)
    conn.execute("call threads=1")
    conn.execute("call auto_checkpoint=false")
    conn.execute("create node table personA (ID INt64, fName StRING, PRIMARY KEY (ID));")
    conn.execute("create node table personB (ID INt64, fName StRING, PRIMARY KEY (ID));")
    conn.execute("create rel table likes (FROM personA TO personB, FROM personB To personA, date DATE);")
    conn.execute(f'COPY personA FROM "{KUZU_ROOT}/dataset/rel-group/node.csv";')
    conn.execute(f'COPY personB FROM "{KUZU_ROOT}/dataset/rel-group/node.csv";')
    conn.execute(f'COPY likes FROM "{KUZU_ROOT}/dataset/rel-group/edge.csv" (FROM="personA", TO="personB");')
    conn.execute(f'COPY likes FROM "{KUZU_ROOT}/dataset/rel-group/edge.csv" (FROM="personB", TO="personA");')
    return db, conn


def test_fsm_reclaim_list_column(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person", "workedHours")
    conn.execute("alter table person drop workedHours")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_string_column(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person", "fName")
    conn.execute("alter table person drop fName")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_list_of_lists(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person", "courseScoresPerTerm")
    conn.execute("alter table person drop courseScoresPerTerm")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_node_table(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person")
    conn.execute("drop table person")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_node_table_recopy(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    prev_num_pages = get_total_used_pages(conn)
    conn.execute("drop table person")
    conn.execute("checkpoint")

    conn.execute(
        "create node table person (ID INt64, fName StRING, gender INT64, isStudent BoOLEAN, isWorker BOOLEAN, age INT64, eyeSight DOUBLE, birthdate DATE, registerTime TIMESTAMP, lastJobDuration interval, workedHours INT64[], usedNames STRING[], courseScoresPerTerm INT64[][], grades INT64[4], height float, u UUID, PRIMARY KEY (ID));"
    )
    conn.execute(
        f"COPY person FROM ['{KUZU_ROOT}/dataset/tinysnb/vPerson.csv', '{KUZU_ROOT}/dataset/tinysnb/vPerson2.csv'](ignore_errors=true, header=false)"
    )
    new_num_pages = get_total_used_pages(conn)
    assert prev_num_pages == new_num_pages


def test_fsm_reclaim_node_table_delete(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person")
    conn.execute("match (p:person) delete p")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_rel_table(fsm_rel_table_setup) -> None:
    _, conn = fsm_rel_table_setup
    used_pages = get_used_page_ranges(conn, "knows")
    conn.execute("drop table knows")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_rel_table_delete(fsm_node_table_setup) -> None:
    _, conn = fsm_node_table_setup
    used_pages = get_used_page_ranges(conn, "person")
    conn.execute("match (p:person) delete p")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_struct(fsm_rel_table_setup) -> None:
    _, conn = fsm_rel_table_setup
    used_pages = get_used_page_ranges(conn, "knows", "fwd_summary") + get_used_page_ranges(
        conn, "knows", "bwd_summary"
    )
    conn.execute("alter table knows drop summary")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_rel_group(fsm_rel_group_setup) -> None:
    _, conn = fsm_rel_group_setup
    used_pages = get_used_page_ranges(conn, "likes")
    conn.execute("drop table likes")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)


def test_fsm_reclaim_rel_group_column(fsm_rel_group_setup) -> None:
    _, conn = fsm_rel_group_setup
    used_pages = get_used_page_ranges(conn, "likes", "fwd_date") + get_used_page_ranges(conn, "likes", "bwd_date")
    conn.execute("alter table likes drop date")
    prevent_data_file_truncation(conn)
    conn.execute("checkpoint")
    free_pages = get_free_page_ranges(conn)
    compare_page_range_lists(used_pages, free_pages)
